# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import numpy as np
import gradio as gr
from PIL import Image, ImageFilter, ImageOps
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import is_opencv_available
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.webuis import (
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
)
from unitorch.cli.webuis import SimpleWebUI


class CannyWebUI(SimpleWebUI):
    def __init__(self, config: CoreConfigureParser):
        # create elements
        input_image = create_element("image", "Input Image")
        width = create_element(
            "slider", "Width", min_value=0, max_value=1024, step=1, default=0
        )
        height = create_element(
            "slider", "Height", min_value=0, max_value=1024, step=1, default=0
        )
        generate = create_element("button", "Generate")
        output_image = create_element("image", "Output Image")

        # create layouts
        left = create_column(input_image, width, height, generate)
        right = create_column(output_image)
        iface = create_blocks(create_row(left, right))

        iface.__enter__()

        generate.click(
            fn=self.canny, inputs=[input_image, width, height], outputs=[output_image]
        )

        iface.__exit__()
        super().__init__(config, iname="Canny", iface=iface)

    def canny(
        self,
        image: Image.Image,
        width: Optional[int] = 0,
        height: Optional[int] = 0,
    ):
        if width > 0 and height > 0:
            image = image.resize((width, height))

        if is_opencv_available():
            import cv2

            image = np.array(image, np.uint8)
            image = cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
            image = cv2.Canny(image, 100, 200)
            image = Image.fromarray(image)
        else:
            image = image.convert("L")
            image = image.filter(ImageFilter.FIND_EDGES)
        return image


class BlendWebUI(SimpleWebUI):
    def __init__(self, config: CoreConfigureParser):
        # create elements
        input_image1 = create_element("image", "Input Image1")
        input_image2 = create_element("image", "Input Image2")
        alpha = create_element(
            "slider", "Alpha", min_value=0, max_value=1, step=0.01, default=0.5
        )
        height = create_element(
            "slider", "Height", min_value=0, max_value=1024, step=1, default=0
        )
        width = create_element(
            "slider", "Width", min_value=0, max_value=1024, step=1, default=0
        )
        generate = create_element("button", "Generate")
        output_image = create_element("image", "Output Image")

        # create layouts
        left = create_column(
            create_row(input_image1, input_image2), height, width, alpha, generate
        )
        right = create_column(output_image)
        iface = create_blocks(create_row(left, right))

        # create events
        iface.__enter__()

        generate.click(
            fn=self.blend,
            inputs=[input_image1, input_image2, alpha, height, width],
            outputs=[output_image],
        )

        iface.__exit__()
        super().__init__(config, iname="Blend", iface=iface)

    def blend(
        self,
        image1,
        image2,
        alpha: Optional[float] = 0.5,
        height: Optional[int] = 0,
        width: Optional[int] = 0,
    ):
        if height > 0 and width > 0:
            image1 = image1.resize((width, height))
            image2 = image2.resize((width, height))
        image = Image.blend(image1, image2, alpha)
        return image


class InvertWebUI(SimpleWebUI):
    def __init__(self, config: CoreConfigureParser):
        # create elements
        input_image = create_element("image", "Input Image")
        height = create_element(
            "slider", "Height", min_value=0, max_value=1024, step=1, default=0
        )
        width = create_element(
            "slider", "Width", min_value=0, max_value=1024, step=1, default=0
        )
        generate = create_element("button", "Generate")
        output_image = create_element("image", "Output Image")

        # create layouts
        left = create_column(input_image, height, width, generate)
        right = create_column(output_image)
        iface = create_blocks(create_row(left, right))

        # create events
        iface.__enter__()

        generate.click(
            fn=self.invert, inputs=[input_image, height, width], outputs=[output_image]
        )

        iface.__exit__()

        super().__init__(config, iname="Invert", iface=iface)

    def invert(
        self,
        image,
        height: Optional[int] = 0,
        width: Optional[int] = 0,
    ):
        if height > 0 and width > 0:
            image = image.resize((width, height))
        image = ImageOps.invert(image)
        return image


class CompositeWebUI(SimpleWebUI):
    def __init__(self, config: CoreConfigureParser):
        input_image = create_element("image", "Input Image")
        input_mask = create_element("image", "Input Mask")
        height = create_element(
            "slider", "Height", min_value=0, max_value=1024, step=1, default=0
        )
        width = create_element(
            "slider", "Width", min_value=0, max_value=1024, step=1, default=0
        )
        generate = create_element("button", "Generate")
        output_image = create_element("image", "Output Image")

        # create layouts
        left = create_column(
            create_row(input_image, input_mask), height, width, generate
        )
        right = create_column(output_image)
        iface = create_blocks(create_row(left, right))

        # create events
        iface.__enter__()

        generate.click(
            fn=self.composite,
            inputs=[input_image, input_mask, height, width],
            outputs=[output_image],
        )

        iface.__exit__()
        super().__init__(config, iname="Composite", iface=iface)

    def composite(
        self,
        image,
        mask,
        height: Optional[int] = 0,
        width: Optional[int] = 0,
    ):
        if height > 0 and width > 0:
            image = image.resize((width, height))
            mask = mask.resize((width, height))
        mask = mask.convert("L")
        result = Image.new("RGBA", image.size, (0, 0, 0, 0))
        result.paste(image, (0, 0), mask)
        return result


class ImageWebUI(SimpleWebUI):
    def __init__(self, config: CoreConfigureParser):
        webuis = [
            CannyWebUI(config),
            BlendWebUI(config),
            InvertWebUI(config),
            CompositeWebUI(config),
        ]
        iface = gr.TabbedInterface(
            [webui.iface for webui in webuis],
            tab_names=[webui.iname for webui in webuis],
        )
        super().__init__(config, iname="Image", iface=iface)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import random
from functools import partial
from random import randint, shuffle, choice

import numpy as np
import torch
from transformers import BertTokenizer

from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import pop_value, truncate_sequence_pair
from unitorch.models import HfTextClassificationProcessor, GenericOutputs


def get_random_word(vocab_words: List[str]) -> str:
    i = random.randint(0, len(vocab_words) - 1)
    return vocab_words[i]


def get_random_mask_indexes(
    tokens: List[str],
    masked_lm_prob: Optional[float] = 0.15,
    do_whole_word_mask: Optional[bool] = True,
    max_predictions_per_seq: Optional[int] = 20,
    special_tokens: List[str] = [],
) -> List[int]:
    cand_indexes = []
    for i, token in enumerate(tokens):
        if token in special_tokens:
            continue
        if (
            do_whole_word_mask
            and len(cand_indexes) >= 1
            and token.startswith("##")
            and cand_indexes[-1][-1] == i - 1
        ):
            cand_indexes[-1].append(i)
        else:
            cand_indexes.append([i])
    random.shuffle(cand_indexes)
    num_to_predict = min(
        max_predictions_per_seq, max(1, int(round(len(tokens) * masked_lm_prob)))
    )
    covered_indexes = set()
    for index_set in cand_indexes:
        if len(covered_indexes) >= num_to_predict:
            break
        if len(covered_indexes) + len(index_set) > num_to_predict or any(
            i in covered_indexes for i in index_set
        ):
            continue
        covered_indexes.update(index_set)
    return list(covered_indexes)


def get_bert_tokenizer(
    vocab_path,
    do_lower_case: Optional[bool] = True,
    do_basic_tokenize: Optional[bool] = True,
    special_input_ids: Optional[Dict] = dict(),
):
    assert os.path.exists(vocab_path)
    tokenizer = BertTokenizer(
        vocab_path,
        do_lower_case=do_lower_case,
        do_basic_tokenize=do_basic_tokenize,
    )
    for token, _id in special_input_ids.items():
        tokenizer.added_tokens_encoder[token] = _id
        tokenizer.unique_no_split_tokens.append(token)
        tokenizer.added_tokens_decoder[_id] = token
        tokenizer.add_tokens(token, special_tokens=True)
    return tokenizer


class BertProcessor(HfTextClassificationProcessor):
    def __init__(
        self,
        vocab_path: str,
        max_seq_length: Optional[int] = 128,
        special_input_ids: Optional[Dict] = dict(),
        do_lower_case: Optional[bool] = True,
        do_basic_tokenize: Optional[bool] = True,
        do_whole_word_mask: Optional[bool] = True,
        masked_lm_prob: Optional[float] = 0.15,
        max_predictions_per_seq: Optional[int] = 20,
    ):
        """
        Initializes the BertProcessor.

        Args:
            vocab_path (str): The path to the vocabulary file.
            max_seq_length (Optional[int], optional): The maximum sequence length. Defaults to 128.
            special_input_ids (Optional[Dict], optional): Special input IDs mapping. Defaults to an empty dictionary.
            do_lower_case (Optional[bool], optional): Whether to perform lowercase tokenization. Defaults to True.
            do_basic_tokenize (Optional[bool], optional): Whether to perform basic tokenization. Defaults to True.
            do_whole_word_mask (Optional[bool], optional): Whether to perform whole word masking. Defaults to True.
            masked_lm_prob (Optional[float], optional): The probability of masking a token for pretraining. Defaults to 0.15.
            max_predictions_per_seq (Optional[int], optional): The maximum number of masked tokens per sequence for pretraining. Defaults to 20.
        """
        tokenizer = get_bert_tokenizer(
            vocab_path,
            do_lower_case=do_lower_case,
            do_basic_tokenize=do_basic_tokenize,
            special_input_ids=special_input_ids,
        )
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
        )
        self.do_whole_word_mask = do_whole_word_mask
        self.masked_lm_prob = masked_lm_prob
        self.max_predictions_per_seq = max_predictions_per_seq
        self.vocab_words = list(self.tokenizer.vocab.keys())

    def pretrain(
        self,
        text: str,
        text_pair: str,
        nsp_label: int,
        max_seq_length: Optional[int] = None,
        masked_lm_prob: Optional[float] = None,
        do_whole_word_mask: Optional[bool] = None,
        max_predictions_per_seq: Optional[int] = None,
    ):
        """
        The Bert pretrain processor on the given text and text pair.

        Args:
            text (str): The input text.
            text_pair (str): The input text pair.
            nsp_label (int): The next sentence prediction label.
            max_seq_length (Optional[int], optional): The maximum sequence length. Defaults to None.
            masked_lm_prob (Optional[float], optional): The probability of masking a token for pretraining. Defaults to None.
            do_whole_word_mask (Optional[bool], optional): Whether to perform whole word masking. Defaults to None.
            max_predictions_per_seq (Optional[int], optional): The maximum number of masked tokens per sequence for pretraining. Defaults to None.

        Returns:
            GenericOutputs: pretrain processing outputs.
        """
        max_seq_length = pop_value(
            max_seq_length,
            self.max_seq_length,
        )

        masked_lm_prob = pop_value(
            masked_lm_prob,
            self.masked_lm_prob,
        )

        do_whole_word_mask = pop_value(
            do_whole_word_mask,
            self.do_whole_word_mask,
        )

        max_predictions_per_seq = pop_value(
            max_predictions_per_seq,
            self.max_predictions_per_seq,
        )

        _tokens = self.tokenizer.tokenize(str(text))
        tokens_pair = self.tokenizer.tokenize(str(text_pair))
        truncate_sequence_pair(_tokens, tokens_pair, max_seq_length - 3)
        tokens = (
            [self.cls_token]
            + _tokens
            + [self.sep_token]
            + tokens_pair
            + [self.sep_token]
        )

        covered_indexes = get_random_mask_indexes(
            tokens,
            masked_lm_prob,
            do_whole_word_mask,
            max_predictions_per_seq,
            special_tokens=[self.cls_token, self.sep_token],
        )
        label = [
            tokens[pos] if pos in covered_indexes else self.pad_token
            for pos in range(max_seq_length)
        ]
        label_mask = [
            1 if pos in covered_indexes else 0 for pos in range(max_seq_length)
        ]
        label = self.tokenizer.convert_tokens_to_ids(label)

        for index in covered_indexes:
            mask_token = None
            if random.random() < 0.8:
                mask_token = self.mask_token
            else:
                mask_token = (
                    tokens[index]
                    if random.random() < 0.5
                    else get_random_word(self.vocab_words)
                )
            tokens[index] = mask_token

        input_ids = self.tokenizer.convert_tokens_to_ids(tokens)
        token_type_ids = [0] + [0] * len(_tokens) + [0] + [1] * len(tokens_pair) + [1]
        attention_mask = [1] * len(input_ids)

        padding = [0] * (max_seq_length - len(input_ids))
        input_ids += len(padding) * [self.pad_token_id]
        attention_mask += padding
        token_type_ids += len(padding) * [1]

        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        return GenericOutputs(
            input_ids=torch.tensor(input_ids, dtype=torch.long),
            token_type_ids=torch.tensor(token_type_ids, dtype=torch.long),
            attention_mask=torch.tensor(attention_mask, dtype=torch.long),
            position_ids=torch.tensor(list(range(max_seq_length)), dtype=torch.long),
            nsp_label=torch.tensor(int(nsp_label), dtype=torch.long),
            mlm_label=torch.tensor(label, dtype=torch.long),
            mlm_label_mask=torch.tensor(label_mask, dtype=torch.long),
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.visualbert import VisualBertProcessor as _VisualBertProcessor
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import WriterOutputs
from unitorch.cli.models import (
    TensorsInputs,
    GenerationOutputs,
    GenerationTargets,
)
from unitorch.cli.models.visualbert import pretrained_visualbert_infos


class VisualBertProcessor(_VisualBertProcessor):
    """VisualBERT Processor for text and image inputs."""

    def __init__(
        self,
        vocab_path,
        max_seq_length: Optional[int] = 128,
        special_input_ids: Optional[Dict] = dict(),
        do_lower_case: Optional[bool] = True,
        do_basic_tokenize: Optional[bool] = True,
        do_whole_word_mask: Optional[bool] = True,
        masked_lm_prob: Optional[float] = 0.15,
        max_predictions_per_seq: Optional[int] = 20,
    ):
        """
        Initialize the VisualBertProcessor.

        Args:
            vocab_path (str): The path to the vocabulary file.
            max_seq_length (Optional[int]): The maximum sequence length. Defaults to 128.
            special_input_ids (Optional[Dict]): A dictionary containing special input IDs. Defaults to an empty dictionary.
            do_lower_case (Optional[bool]): Whether to convert the text to lowercase. Defaults to True.
            do_basic_tokenize (Optional[bool]): Whether to perform basic tokenization. Defaults to True.
            do_whole_word_mask (Optional[bool]): Whether to use whole-word masking. Defaults to True.
            masked_lm_prob (Optional[float]): The probability of masked language model masking. Defaults to 0.15.
            max_predictions_per_seq (Optional[int]): The maximum number of masked language model predictions per sequence. Defaults to 20.
        """
        super().__init__(
            vocab_path=vocab_path,
            max_seq_length=max_seq_length,
            special_input_ids=special_input_ids,
            do_lower_case=do_lower_case,
            do_basic_tokenize=do_basic_tokenize,
            do_whole_word_mask=do_whole_word_mask,
            masked_lm_prob=masked_lm_prob,
            max_predictions_per_seq=max_predictions_per_seq,
        )

    @classmethod
    @add_default_section_for_init("core/process/visualbert")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of VisualBertProcessor from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            dict: A dictionary containing the processor's configuration.
        """
        config.set_default_section("core/process/visualbert")
        pretrained_name = config.getoption("pretrained_name", "default-visualbert")
        vocab_path = config.getoption("vocab_path", None)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_visualbert_infos, pretrained_name, "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        return {
            "vocab_path": vocab_path,
        }

    @register_process("core/process/visualbert/classification")
    def _classification(
        self,
        text: str,
        text_pair: Optional[str] = None,
        max_seq_length: Optional[int] = None,
    ):
        """
        Perform classification processing on the input text.

        Args:
            text (str): The input text.
            text_pair (Optional[str]): The second input text for sentence pair classification. Defaults to None.
            max_seq_length (Optional[int]): The maximum sequence length. Defaults to None.

        Returns:
            TensorsInputs: The processed tensors as inputs to the model.
        """
        outputs = super().classification(
            text=text,
            text_pair=text_pair,
            max_seq_length=max_seq_length,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            token_type_ids=outputs.token_type_ids,
            position_ids=outputs.position_ids,
        )

    @register_process("core/process/visualbert/pretrain")
    def _pretrain(
        self,
        text: str,
        text_pair: str,
        nsp_label: int,
        max_seq_length: Optional[int] = None,
        masked_lm_prob: Optional[float] = None,
        do_whole_word_mask: Optional[bool] = None,
        max_predictions_per_seq: Optional[int] = None,
    ):
        """
        Perform pretraining processing on the input text and labels.

        Args:
            text (str): The input text.
            text_pair (str): The second input text for sentence pair pretraining.
            nsp_label (int): The next sentence prediction label.
            max_seq_length (Optional[int]): The maximum sequence length. Defaults to None.
            masked_lm_prob (Optional[float]): The probability of masked language model masking. Defaults to None.
            do_whole_word_mask (Optional[bool]): Whether to use whole-word masking. Defaults to None.
            max_predictions_per_seq (Optional[int]): The maximum number of masked language model predictions per sequence. Defaults to None.

        Returns:
            TensorsInputs: The processed tensors as inputs to the model.
        """
        outputs = super().pretrain(
            text=text,
            text_pair=text_pair,
            nsp_label=nsp_label,
            max_seq_length=max_seq_length,
            masked_lm_prob=masked_lm_prob,
            do_whole_word_mask=do_whole_word_mask,
            max_predictions_per_seq=max_predictions_per_seq,
        )
        return TensorsInputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            token_type_ids=outputs.token_type_ids,
            position_ids=outputs.position_ids,
            nsp_label=outputs.nsp_label,
            mlm_label=outputs.mlm_label,
            mlm_label_mask=outputs.mlm_label_mask,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_llama_infos = {
    # llama
    "default-llama": {
        "config": "https://huggingface.co/huggyllama/llama-7b/resolve/main/config.json",
        "vocab": "https://huggingface.co/huggyllama/llama-7b/resolve/main/tokenizer.model",
    },
    "llama-7b": {
        "config": "https://huggingface.co/huggyllama/llama-7b/resolve/main/config.json",
        "vocab": "https://huggingface.co/huggyllama/llama-7b/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/huggyllama/llama-7b/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00002.bin"
            for i in range(1, 3)
        ],
    },
    "llama-13b": {
        "config": "https://huggingface.co/huggyllama/llama-13b/resolve/main/config.json",
        "vocab": "https://huggingface.co/huggyllama/llama-13b/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/huggyllama/llama-13b/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00003.bin"
            for i in range(1, 4)
        ],
    },
    "llama2-7b": {
        "config": "https://huggingface.co/NousResearch/Llama-2-7b-hf/resolve/main/config.json",
        "vocab": "https://huggingface.co/NousResearch/Llama-2-7b-hf/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/NousResearch/Llama-2-7b-hf/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00003.bin"
            for i in range(1, 4)
        ],
    },
    "llama2-chat-7b": {
        "config": "https://huggingface.co/NousResearch/Llama-2-7b-chat-hf/resolve/main/config.json",
        "vocab": "https://huggingface.co/NousResearch/Llama-2-7b-chat-hf/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/NousResearch/Llama-2-7b-chat-hf/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00003.bin"
            for i in range(1, 4)
        ],
    },
    "llama2-13b": {
        "config": "https://huggingface.co/NousResearch/Llama-2-13b-hf/resolve/main/config.json",
        "vocab": "https://huggingface.co/NousResearch/Llama-2-13b-hf/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/NousResearch/Llama-2-13b-hf/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00006.bin"
            for i in range(1, 7)
        ],
    },
    "llama2-chat-13b": {
        "config": "https://huggingface.co/NousResearch/Llama-2-13b-chat-hf/resolve/main/config.json",
        "vocab": "https://huggingface.co/NousResearch/Llama-2-13b-chat-hf/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/NousResearch/Llama-2-13b-chat-hf/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00003.bin"
            for i in range(1, 4)
        ],
    },
    "open-llama-3b": {
        "config": "https://huggingface.co/openlm-research/open_llama_3b/resolve/main/config.json",
        "vocab": "https://huggingface.co/openlm-research/open_llama_3b/resolve/main/tokenizer.model",
        "weight": "https://huggingface.co/openlm-research/open_llama_3b/resolve/main/pytorch_model.bin",
    },
    "open-llama-7b": {
        "config": "https://huggingface.co/openlm-research/open_llama_7b/resolve/main/config.json",
        "vocab": "https://huggingface.co/openlm-research/open_llama_7b/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/openlm-research/open_llama_7b/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00002.bin"
            for i in range(1, 3)
        ],
    },
    "open-llama-13b": {
        "config": "https://huggingface.co/openlm-research/open_llama_13b/resolve/main/config.json",
        "vocab": "https://huggingface.co/openlm-research/open_llama_13b/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/openlm-research/open_llama_13b/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00003.bin"
            for i in range(1, 4)
        ],
    },
    "llama3-8b-instruct": {
        "config": "https://huggingface.co/unsloth/llama-3-8b-Instruct/resolve/main/config.json",
        "vocab": "https://huggingface.co/unsloth/llama-3-8b-Instruct/resolve/main/tokenizer.model",
        "weight": [
            f"https://huggingface.co/unsloth/llama-3-8b-Instruct/resolve/main/pytorch_model-{str(i).rjust(5, '0')}-of-00004.bin"
            for i in range(1, 5)
        ],
    },
}

pretrained_llama_extensions_infos = {}

import unitorch.cli.models.llama.modeling
import unitorch.cli.models.llama.processing
from unitorch.cli.models.llama.modeling import (
    LlamaForClassification,
    LlamaForGeneration,
)
from unitorch.cli.models.llama.processing import LlamaProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.deberta.modeling import DebertaForClassification, DebertaForMaskLM
from unitorch.models.deberta.processing import DebertaProcessor, get_deberta_tokenizer
from unitorch.models.deberta.modeling_v2 import DebertaV2ForClassification
from unitorch.models.deberta.processing_v2 import (
    DebertaV2Processor,
    get_deberta_v2_tokenizer,
)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_sam_infos = {
    "default-sam": {
        "config": "https://huggingface.co/facebook/sam-vit-base/resolve/main/config.json",
        "vision_config": "https://huggingface.co/facebook/sam-vit-base/resolve/main/preprocessor_config.json",
    },
    "sam-vit-base": {
        "config": "https://huggingface.co/facebook/sam-vit-base/resolve/main/config.json",
        "vision_config": "https://huggingface.co/facebook/sam-vit-base/resolve/main/preprocessor_config.json",
        "weight": "https://huggingface.co/facebook/sam-vit-base/resolve/main/pytorch_model.bin",
    },
    "sam-vit-large": {
        "config": "https://huggingface.co/facebook/sam-vit-large/resolve/main/config.json",
        "vision_config": "https://huggingface.co/facebook/sam-vit-large/resolve/main/preprocessor_config.json",
        "weight": "https://huggingface.co/facebook/sam-vit-large/resolve/main/pytorch_model.bin",
    },
    "sam-vit-huge": {
        "config": "https://huggingface.co/facebook/sam-vit-huge/resolve/main/config.json",
        "vision_config": "https://huggingface.co/facebook/sam-vit-huge/resolve/main/preprocessor_config.json",
        "weight": "https://huggingface.co/facebook/sam-vit-huge/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.sam.modeling
import unitorch.cli.models.sam.processing
from unitorch.cli.models.sam.modeling import SamForSegmentation
from unitorch.cli.models.sam.processing import SamProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import json
import logging
import torch
import torch.nn as nn
import torch.nn.functional as F
import transformers
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from transformers import BloomModel, BloomConfig, BloomForCausalLM
from unitorch.utils.decorators import replace
from unitorch.models import GenericModel, GenericOutputs
from unitorch.models.peft import PeftWeightLoaderMixin


class BloomForClassification(GenericModel, PeftWeightLoaderMixin):
    """
    A classification model based on the Bloom architecture.
    """

    prefix_keys_in_state_dict = {"^(?!transformer\.).*": "transformer."}
    replace_keys_in_peft_state_dict = {"peft_model.base_model.model.": "transformer."}

    def __init__(
        self,
        config_path: str,
        num_classes: Optional[int] = 1,
        hidden_dropout_prob: Optional[float] = 0.1,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes a new instance of the BloomForClassification model.

        Args:
            config_path (str): The path to the configuration file for the Bloom model.
            num_classes (Optional[int]): The number of output classes for classification. Defaults to 1.
            hidden_dropout_prob (Optional[float]): The dropout probability for the hidden layers. Defaults to 0.1.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = BloomConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.transformer = BloomModel(self.config)
        self.dropout = nn.Dropout(hidden_dropout_prob)
        self.classifier = nn.Linear(self.config.hidden_size, num_classes)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Forward pass of the BloomForClassification model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            attention_mask (Optional[torch.Tensor]): The attention mask tensor. Defaults to None.

        Returns:
            (torch.Tensor):The output logits for classification.
        """
        outputs = self.transformer(
            input_ids,
            attention_mask=attention_mask,
        )[0]
        pooled_output = outputs[:, -1]
        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)
        return logits


class BloomForGeneration(GenericModel, PeftWeightLoaderMixin):
    """
    A generation model based on the Bloom architecture.
    """

    prefix_keys_in_state_dict = {"^(?!model\.).*": "model.transformer."}
    replace_keys_in_peft_state_dict = {"peft_model.base_model.": ""}

    def __init__(
        self,
        config_path: str,
        gradient_checkpointing: Optional[bool] = False,
    ):
        """
        Initializes a new instance of the BloomForGeneration model.

        Args:
            config_path (str): The path to the configuration file for the Bloom model.
            gradient_checkpointing (Optional[bool]): Whether to use gradient checkpointing. Defaults to False.
        """
        super().__init__()
        self.config = BloomConfig.from_json_file(config_path)
        self.config.gradient_checkpointing = gradient_checkpointing
        self.model = BloomForCausalLM(self.config)
        self.init_weights()

    def forward(
        self,
        input_ids: torch.Tensor,
        attention_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        Forward pass of the BloomForGeneration model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            attention_mask (Optional[torch.Tensor]): The attention mask tensor. Defaults to None.

        Returns:
            (torch.Tensor):The output logits.
        """
        outputs = self.model(
            input_ids,
            attention_mask=attention_mask,
            return_dict=True,
        )
        logits = outputs.logits
        return logits

    @torch.no_grad()
    def generate(
        self,
        input_ids: torch.Tensor,
        num_beams: Optional[int] = 5,
        decoder_start_token_id: Optional[int] = 1,
        decoder_end_token_id: Optional[Union[int, List[int]]] = 2,
        num_return_sequences: Optional[int] = 1,
        min_gen_seq_length: Optional[int] = 0,
        max_gen_seq_length: Optional[int] = 48,
        repetition_penalty: Optional[float] = 1.0,
        no_repeat_ngram_size: Optional[int] = 0,
        early_stopping: Optional[bool] = True,
        length_penalty: Optional[float] = 1.0,
        num_beam_groups: Optional[int] = 1,
        diversity_penalty: Optional[float] = 0.0,
        do_sample: Optional[bool] = False,
        temperature: Optional[float] = 1.0,
        top_k: Optional[int] = 50,
        top_p: Optional[float] = 1.0,
    ) -> GenericOutputs:
        """
        Generate sequences using the BloomForGeneration model.

        Args:
            input_ids (torch.Tensor): The input token IDs.
            num_beams (Optional[int]): The number of beams for beam search. Defaults to 5.
            decoder_start_token_id (Optional[int]): The ID of the start token for decoding. Defaults to 1.
            decoder_end_token_id (Optional[int]): The ID of the end token for decoding. Defaults to 2.
            num_return_sequences (Optional[int]): The number of generated sequences to return. Defaults to 1.
            min_gen_seq_length (Optional[int]): The minimum length of the generated sequences. Defaults to 0.
            max_gen_seq_length (Optional[int]): The maximum length of the generated sequences. Defaults to 48.
            repetition_penalty (Optional[float]): The penalty for repeated n-grams. Defaults to 1.0.
            no_repeat_ngram_size (Optional[int]): The size of n-grams to prevent repetition. Defaults to 0.
            early_stopping (Optional[bool]): Whether to stop generation early based on specified conditions. Defaults to True.
            length_penalty (Optional[float]): The penalty for longer sequences. Defaults to 1.0.
            num_beam_groups (Optional[int]): The number of beam groups for diverse beam search. Defaults to 1.
            diversity_penalty (Optional[float]): The penalty for diverse beam search. Defaults to 0.0.
            do_sample (Optional[bool]): Whether to use sampling for generation. Defaults to False.
            temperature (Optional[float]): The temperature for sampling. Defaults to 1.0.
            top_k (Optional[int]): The number of top-k tokens to consider for sampling. Defaults to 50.
            top_p (Optional[float]): The cumulative probability for top-p sampling. Defaults to 1.0.

        Returns:
            GenericOutputs: The generated sequences and their scores.
        """
        input_seq_length = input_ids.size(1)
        outputs = self.model.generate(
            input_ids,
            max_length=max_gen_seq_length + input_seq_length,
            min_length=min_gen_seq_length + input_seq_length,
            num_beams=num_beams,
            do_sample=do_sample,
            decoder_start_token_id=decoder_start_token_id,
            no_repeat_ngram_size=no_repeat_ngram_size,
            early_stopping=early_stopping,
            length_penalty=length_penalty,
            repetition_penalty=repetition_penalty,
            num_return_sequences=num_return_sequences,
            bos_token_id=decoder_start_token_id,
            eos_token_id=decoder_end_token_id,
            num_beam_groups=num_beam_groups,
            diversity_penalty=diversity_penalty,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            return_dict_in_generate=True,
            output_scores=True,
        )
        sequences = outputs.sequences.reshape(
            -1, num_return_sequences, outputs.sequences.size(-1)
        )
        outputs.sequences = torch.zeros(
            sequences.size(0), num_return_sequences, max_gen_seq_length
        ).to(device=sequences.device)
        outputs.sequences[:, :, : sequences.size(-1) - input_seq_length].copy_(
            sequences[:, :, input_seq_length : sequences.size(-1)]
        )

        if num_return_sequences == 1:
            outputs.sequences = outputs.sequences.reshape(-1, max_gen_seq_length)

        return GenericOutputs(
            sequences=outputs.sequences.long(),
            sequences_scores=outputs.sequences_scores,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch.cuda.amp import autocast
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.sam import (
    SamForSegmentation as _SamForSegmentation,
    SamProcessor,
)
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import SegmentationOutputs, LossOutputs
from unitorch.cli.models import segmentation_model_decorator
from unitorch.cli.models.sam import pretrained_sam_infos


@register_model("core/model/segmentation/sam", segmentation_model_decorator)
class SamForSegmentation(_SamForSegmentation):
    def __init__(
        self,
        config_path: str,
        vision_config_path: str,
        mask_threshold: Optional[float] = 0.0,
        pred_iou_thresh: Optional[float] = 0.88,
        stability_score_thresh: Optional[float] = 0.95,
        stability_score_offset: Optional[int] = 1,
        crops_nms_thresh: Optional[float] = 0.7,
    ):
        super().__init__(
            config_path=config_path,
        )
        self.processor = SamProcessor(vision_config_path=vision_config_path)
        self.mask_threshold = mask_threshold
        self.pred_iou_thresh = pred_iou_thresh
        self.stability_score_thresh = stability_score_thresh
        self.stability_score_offset = stability_score_offset
        self.crops_nms_thresh = crops_nms_thresh

    @classmethod
    @add_default_section_for_init("core/model/segmentation/sam")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/segmentation/sam")
        pretrained_name = config.getoption("pretrained_name", "default-sam")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_sam_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)

        vision_config_path = config.getoption("vision_config_path", None)
        vision_config_path = pop_value(
            vision_config_path,
            nested_dict_value(pretrained_sam_infos, pretrained_name, "vision_config"),
        )

        vision_config_path = cached_path(vision_config_path)

        inst = cls(
            config_path=config_path,
            vision_config_path=vision_config_path,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_sam_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(
        self,
    ):
        raise NotImplementedError

    @add_default_section_for_function("core/model/segmentation/sam")
    @torch.no_grad()
    def segment(
        self,
        pixel_values: torch.Tensor,
        input_points: torch.Tensor,
        input_boxes: torch.Tensor,
        original_sizes: torch.Tensor,
        reshaped_input_sizes: torch.Tensor,
        input_labels: Optional[torch.Tensor] = None,
    ):
        outputs = super().segment(
            pixel_values=pixel_values,
            input_points=input_points,
            input_labels=input_labels,
        )
        processed_masks = self.processor.processing_masks(
            masks=outputs.masks,
            scores=outputs.scores,
            input_boxes=input_boxes,
            original_sizes=original_sizes,
            reshaped_input_sizes=reshaped_input_sizes,
            mask_threshold=self.mask_threshold,
            pred_iou_thresh=self.pred_iou_thresh,
            stability_score_thresh=self.stability_score_thresh,
            stability_score_offset=self.stability_score_offset,
            crops_nms_thresh=self.crops_nms_thresh,
        )
        return SegmentationOutputs(
            masks=processed_masks.masks,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os

# env setting
os.environ["TOKENIZERS_PARALLELISM"] = "false"
### cache setting
ENV_VARS_TRUE_VALUES = {"1", "ON", "YES", "TRUE"}
UNITORCH_CACHE = os.getenv(
    "UNITORCH_CACHE", os.path.join(os.getenv("HOME", "."), ".cache/unitorch")
)
os.environ["TRANSFORMERS_CACHE"] = UNITORCH_CACHE
os.environ["HF_DATASETS_CACHE"] = UNITORCH_CACHE


def get_cache_home():
    return UNITORCH_CACHE


### version
VERSION = "0.0.0.12"

### is offline mode
UNITORCH_OFFLINE = os.environ.get("UNITORCH_OFFLINE", "0").upper()


def is_offline_mode():
    return UNITORCH_OFFLINE in ENV_VARS_TRUE_VALUES


### is offline debug mode
UNITORCH_OFFLINE_DEBUG_VALUES = {"1", "ON", "YES", "TRUE", "DEBUG"}
UNITORCH_OFFLINE_DEBUG = os.environ.get("UNITORCH_OFFLINE_DEBUG", "0").upper()


def is_offline_debug_mode():
    return UNITORCH_OFFLINE_DEBUG in UNITORCH_OFFLINE_DEBUG_VALUES


# before setup logging
import sklearn

# logging & warning setting
import logging
import warnings

UNITORCH_DEBUG_VALUES = {"OFF": 50, "INFO": 20, "DETAIL": 10, "CPU": 10, "ALL": 0}
UNITORCH_DEBUG = os.environ.get("UNITORCH_DEBUG", "INFO").upper()
UNITORCH_DEBUG = (
    UNITORCH_DEBUG if UNITORCH_DEBUG in UNITORCH_DEBUG_VALUES.keys() else "INFO"
)
UNITORCH_DEBUG_INT = UNITORCH_DEBUG_VALUES.get(UNITORCH_DEBUG)

if UNITORCH_DEBUG == "CPU":
    os.environ["CUDA_VISIBLE_DEVICES"] = "999"

warnings.filterwarnings("ignore")
logging.basicConfig(
    format="%(asctime)s : %(levelname)s : %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S %p",
    level=logging.getLevelName(UNITORCH_DEBUG_INT),
)

# settings
import torch
import random
import numpy as np
import transformers
import torch.multiprocessing

torch.multiprocessing.set_sharing_strategy("file_system")

# torch.backends.cuda.matmul.allow_tf32 = False
torch.backends.cudnn.deterministic = False
torch.backends.cudnn.benchmark = True
# torch.backends.cudnn.allow_tf32 = True


def set_seed(seed: int):
    """
    Helper function for reproducible behavior to set the seed in `random`, `numpy`, `torch` and/or `tf` (if installed).
    Args:
        seed (`int`): The seed to set.
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    transformers.set_seed(seed)

    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


# useful functions
from unitorch.utils import cached_path as hf_cached_path, read_file, read_json_file
from unitorch.utils import (
    is_deepspeed_available,
    is_megatron_available,
    is_diffusers_available,
    is_xformers_available,
    is_opencv_available,
    is_torch_available,
    is_torch2_available,
)

if is_diffusers_available():
    import diffusers

    diffusers.logging.set_verbosity_error()

# imports from other files
import unitorch.datasets
import unitorch.losses
import unitorch.scores
import unitorch.models
import unitorch.modules
import unitorch.optims
import unitorch.schedulers
import unitorch.scores
import unitorch.tasks

# more classes



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import sys
import fire
import logging
import importlib
import gradio as gr
import importlib_resources
import unitorch.cli
from torch.multiprocessing import spawn
from transformers.utils import is_remote_url
from unitorch.utils import read_file
from unitorch.cli import CoreConfigureParser
from unitorch.cli import (
    import_library,
    cached_path,
    set_global_config,
    registered_webui,
    init_registered_module,
)
import unitorch.cli.webuis


@fire.decorators.SetParseFn(str)
def webui(config_path: str, **kwargs):
    config_path = cached_path(config_path)

    params = []
    for k, v in kwargs.items():
        if k.count("@") > 0:
            k0 = k.split("@")[0]
            k1 = "@".join(k.split("@")[1:])
        else:
            k0 = "core/cli"
            k1 = k
        params.append((k0, k1, v))

    if config_path is not None:
        config = CoreConfigureParser(config_path, params=params)
    else:
        config = CoreConfigureParser(params=params)

    set_global_config(config)

    depends_libraries = config.getdefault("core/cli", "depends_libraries", None)

    if depends_libraries:
        for library in depends_libraries:
            import_library(library)

    enabled_webuis = config.getdefault("core/cli", "enabled_webuis", None)
    single_webui = config.getdefault("core/cli", "single_webui", False)
    assert enabled_webuis is not None
    if isinstance(enabled_webuis, str):
        enabled_webuis = [enabled_webuis]
    assert all(enabled_webui in registered_webui for enabled_webui in enabled_webuis)

    webui_instance = lambda webui_name, config: registered_webui.get(webui_name)["obj"](
        config
    )

    webuis = [webui_instance(enabled_webui, config) for enabled_webui in enabled_webuis]

    if single_webui:
        assert (
            len(webuis) == 1
        ), "single_webui can only be used when there is only one webui enabled"
        webuis = webuis[0]
        demo_webui = webuis.iface
    else:
        demo_webui = gr.TabbedInterface(
            interface_list=[webui.iface for webui in webuis],
            tab_names=[webui.iname for webui in webuis],
            title="🔥 Unitorch WebUI",
        )
    demo_webui.title = "Unitorch WebUI"
    demo_webui.theme_css = read_file(
        os.path.join(importlib_resources.files("unitorch"), "cli/assets/style.css")
    )
    demo_webui.css = demo_webui.theme_css

    config.set_default_section("core/cli")
    host = config.getoption("host", "0.0.0.0")
    port = config.getoption("port", 7860)
    share = config.getoption("share", False)
    ssl_keyfile = config.getoption("ssl_keyfile", None)
    ssl_certfile = config.getoption("ssl_certfile", None)
    ssl_verify = config.getoption("ssl_verify", True)
    demo_webui.launch(
        server_name=host,
        server_port=port,
        share=share,
        favicon_path=os.path.join(
            importlib_resources.files("unitorch"), "cli/assets/icon.png"
        ),
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile,
        ssl_verify=ssl_verify,
    )

    os._exit(0)


def cli_main():
    fire.Fire(webui)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
import numpy as np
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torchvision.transforms import Resize, CenterCrop, ToTensor, Normalize, Compose
from transformers import BlipImageProcessor, BertTokenizer
from transformers.image_utils import to_numpy_array, ChannelDimension
from transformers.image_transforms import to_channel_dimension_format
from unitorch.utils import pop_value
from unitorch.models import (
    HfTextClassificationProcessor,
    HfTextGenerationProcessor,
    HfImageClassificationProcessor,
    GenericOutputs,
)


class BlipProcessor(
    HfTextClassificationProcessor,
    HfTextGenerationProcessor,
    HfImageClassificationProcessor,
):
    def __init__(
        self,
        vocab_path: str,
        vision_config_path: str,
        max_seq_length: Optional[int] = 128,
        max_gen_seq_length: Optional[int] = 48,
        position_start_id: Optional[int] = 0,
    ):
        """
        Initializes the BlipProcessor.

        Args:
            vocab_path (str): The path to the vocabulary file.
            vision_config_path (str): The path to the vision configuration file.
            max_seq_length (Optional[int]): The maximum sequence length for text inputs. Defaults to 128.
            max_gen_seq_length (Optional[int]): The maximum sequence length for generated outputs. Defaults to 48.
            position_start_id (Optional[int]): The position start ID. Defaults to 0.
        """
        vision_processor = BlipImageProcessor.from_json_file(vision_config_path)
        HfImageClassificationProcessor.__init__(self, vision_processor=vision_processor)

        tokenizer = BertTokenizer(
            vocab_file=vocab_path,
        )
        tokenizer.bos_token = tokenizer.cls_token
        tokenizer.bos_token_id = tokenizer.cls_token_id
        tokenizer.eos_token = tokenizer.sep_token
        tokenizer.eos_token_id = tokenizer.sep_token_id
        HfTextClassificationProcessor.__init__(
            self,
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            source_type_id=0,
            target_type_id=0,
            position_start_id=position_start_id,
        )

        HfTextGenerationProcessor.__init__(
            self,
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            max_gen_seq_length=max_gen_seq_length,
        )

    def text_classification(
        self,
        text: str,
        max_seq_length: Optional[int] = None,
    ) -> GenericOutputs:
        """
        Performs text classification on the given input text.

        Args:
            text (str): The input text to classify.
            max_seq_length (Optional[int]): The maximum sequence length for the text. If None, the default value from initialization is used.

        Returns:
            GenericOutputs: The outputs of the text classification.
        """
        outputs = HfTextClassificationProcessor.classification(
            self,
            text=text,
            max_seq_length=max_seq_length,
        )
        return GenericOutputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
            position_ids=outputs.position_ids,
        )

    def image_classification(
        self,
        image: Union[Image.Image, str],
    ) -> GenericOutputs:
        """
        Performs image classification on the given input image.

        Args:
            image (PIL.Image.Image): The input image to classify.

        Returns:
            GenericOutputs: The outputs of the image classification.
        """
        outputs = HfImageClassificationProcessor.classification(
            self,
            image=image,
        )

        return GenericOutputs(
            pixel_values=outputs.pixel_values,
        )

    def classification(
        self,
        text: str,
        image: Union[Image.Image, str],
        max_seq_length: Optional[int] = None,
    ) -> GenericOutputs:
        """
        Performs classification using both text and image inputs.

        Args:
            text (str): The input text to classify.
            image (PIL.Image.Image): The input image to classify.
            max_seq_length (Optional[int]): The maximum sequence length for the text. If None, the default value from initialization is used.

        Returns:
            GenericOutputs: The outputs of the classification.
        """
        max_seq_length = pop_value(
            max_seq_length,
            self.max_seq_length,
        )

        text_outputs = self.text_classification(text, max_seq_length)
        pixel_outputs = self.image_classification(image)

        return GenericOutputs(
            input_ids=text_outputs.input_ids,
            attention_mask=text_outputs.attention_mask,
            position_ids=text_outputs.position_ids,
            pixel_values=pixel_outputs.pixel_values,
        )

    def generation_inputs(
        self,
        text: str,
        max_seq_length: Optional[int] = None,
    ) -> GenericOutputs:
        """
        Generate inputs for text generation.

        Args:
            text (str): The input text.
            max_seq_length (int, optional): Maximum sequence length. Defaults to None.

        Returns:
            GenericOutputs: The generated input tokens and attention mask.
        """
        outputs = HfTextGenerationProcessor.generation_inputs(
            self,
            text=text,
            max_seq_length=max_seq_length,
        )
        return GenericOutputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
        )

    def generation_labels(
        self,
        text: str,
        max_gen_seq_length: Optional[int] = None,
    ) -> GenericOutputs:
        """
        Generates labels for text generation based on the given input text.

        Args:
            text (str): The input text for generating labels.
            max_gen_seq_length (Optional[int]): The maximum sequence length for the generated labels. If None, the default value from initialization is used.

        Returns:
            GenericOutputs: The generated labels.
        """
        outputs = HfTextGenerationProcessor.generation_labels(
            self,
            text=text,
            max_gen_seq_length=max_gen_seq_length,
        )
        return GenericOutputs(
            input_ids=outputs.input_ids,
            attention_mask=outputs.attention_mask,
        )

    def generation(
        self,
        text: str,
        image: Union[Image.Image, str],
        max_gen_seq_length: Optional[int] = None,
    ) -> GenericOutputs:
        """
        Generate inputs, labels, and tokens for image to text generation.

        Args:
            text (str): The input text.
            image (Image.Image): The input image to caption.
            max_gen_seq_length (int, optional): Maximum generated sequence length. Defaults to None.

        Returns:
            GenericOutputs: The generated input tokens, attention masks, label tokens, and attention masks.
        """

        max_gen_seq_length = pop_value(max_gen_seq_length, self.max_gen_seq_length)

        tokens = self.generation_inputs(text, max_gen_seq_length)
        pixels = self.image_classification(image)
        labels = self.generation_labels(text, max_gen_seq_length)

        return GenericOutputs(
            input_ids=tokens.input_ids,
            attention_mask=tokens.attention_mask,
            pixel_values=pixels.pixel_values,
            input_ids_label=labels.input_ids,
            attention_mask_label=labels.attention_mask,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import torch
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from transformers import PegasusTokenizer
from unitorch.models import HfTextGenerationProcessor


def get_pegasus_tokenizer(
    vocab_path,
    special_input_ids=dict(),
):
    """
    Creates and returns an instance of the Pegasus tokenizer.

    Args:
        vocab_path (str): The path to the vocabulary file.
        special_input_ids (Dict, optional): Dictionary of special input tokens and their corresponding IDs. Defaults to an empty dictionary.

    Returns:
        PegasusTokenizer: The Pegasus tokenizer.
    """
    tokenizer = PegasusTokenizer(vocab_path)
    for token, _id in special_input_ids.items():
        tokenizer.added_tokens_encoder[token] = _id
        tokenizer.unique_no_split_tokens.append(token)
        tokenizer.added_tokens_decoder[_id] = token
        tokenizer.add_tokens(token, special_tokens=True)
    return tokenizer


class PegasusProcessor(HfTextGenerationProcessor):
    def __init__(
        self,
        vocab_path: str,
        special_input_ids: Optional[Dict] = dict(),
        max_seq_length: Optional[int] = 128,
        max_gen_seq_length: Optional[int] = 48,
    ):
        """
        Initializes a PegasusProcessor.

        Args:
            vocab_path (str): The path to the vocabulary file.
            special_input_ids (Dict, optional): Dictionary of special input tokens and their corresponding IDs. Defaults to an empty dictionary.
            max_seq_length (int, optional): The maximum length of input sequences. Defaults to 128.
            max_gen_seq_length (int, optional): The maximum length of generated sequences. Defaults to 48.
        """
        tokenizer = get_pegasus_tokenizer(
            vocab_path,
            special_input_ids=special_input_ids,
        )
        tokenizer.bos_token_id = 0
        tokenizer.bos_token = tokenizer.convert_ids_to_tokens(0)
        tokenizer.sep_token = tokenizer.eos_token
        tokenizer.sep_token_id = tokenizer.eos_token_id
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            max_gen_seq_length=max_gen_seq_length,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos

pretrained_bert_infos = {
    "default-bert": {
        "config": "https://huggingface.co/bert-base-uncased/resolve/main/config.json",
        "vocab": "https://huggingface.co/bert-base-uncased/resolve/main/vocab.txt",
    },
    "bert-base-uncased": {
        "config": "https://huggingface.co/bert-base-uncased/resolve/main/config.json",
        "vocab": "https://huggingface.co/bert-base-uncased/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/bert-base-uncased/resolve/main/pytorch_model.bin",
    },
    "distilbert-base-uncased": {
        "config": "https://huggingface.co/distilbert-base-uncased/resolve/main/config.json",
        "vocab": "https://huggingface.co/distilbert-base-uncased/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/distilbert-base-uncased/resolve/main/pytorch_model.bin",
    },
    "bert-base-multilingual-uncased": {
        "config": "https://huggingface.co/bert-base-multilingual-uncased/resolve/main/config.json",
        "vocab": "https://huggingface.co/bert-base-multilingual-uncased/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/bert-base-multilingual-uncased/resolve/main/pytorch_model.bin",
    },
    "bert-base-multilingual-cased": {
        "config": "https://huggingface.co/bert-base-multilingual-cased/resolve/main/config.json",
        "vocab": "https://huggingface.co/bert-base-multilingual-cased/resolve/main/vocab.txt",
        "weight": "https://huggingface.co/bert-base-multilingual-cased/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.bert.modeling
import unitorch.cli.models.bert.processing
from unitorch.cli.models.bert.modeling import BertForClassification
from unitorch.cli.models.bert.processing import BertProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
import numpy as np
from sklearn.metrics import (
    accuracy_score,
    f1_score,
    matthews_corrcoef,
    precision_score,
    recall_score,
    roc_auc_score,
    ndcg_score,
    auc,
    precision_recall_curve,
)
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union


def convert_tensor_to_strings(inputs: torch.Tensor) -> Any:
    if inputs.dim() == 0:
        return str(inputs.item())
    elif inputs.dim() == 1:
        return [str(element.item()) for element in inputs]
    else:
        return [convert_tensor_to_strings(tensor) for tensor in inputs]


def remove_strings_ignore_tokens(inputs: Any, ignore_tokens: Optional[Set[str]]) -> Any:
    if ignore_tokens is None:
        return inputs

    if isinstance(inputs, list):
        return [
            remove_strings_ignore_tokens(element, ignore_tokens) for element in inputs
        ]
    elif isinstance(inputs, dict):
        return {
            key: remove_strings_ignore_tokens(value, ignore_tokens)
            for key, value in inputs.items()
        }
    elif isinstance(inputs, tuple):
        return tuple(
            remove_strings_ignore_tokens(element, ignore_tokens) for element in inputs
        )
    elif isinstance(inputs, set):
        return {
            remove_strings_ignore_tokens(element, ignore_tokens) for element in inputs
        }
    elif isinstance(inputs, frozenset):
        return frozenset(
            remove_strings_ignore_tokens(element, ignore_tokens) for element in inputs
        )
    else:
        return inputs


pearsonr = lambda y_true, y_pred: np.corrcoef(y_true, y_pred)[0, 1]

from unitorch.scores.bleu import bleu_score
from unitorch.scores.rouge import rouge1_score, rouge2_score, rougel_score
from unitorch.scores.map import map_score, map50_score



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_bart_infos = {
    "default-bart": {
        "config": "https://huggingface.co/facebook/bart-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/facebook/bart-base/resolve/main/vocab.json",
        "merge": "https://huggingface.co/facebook/bart-base/resolve/main/merges.txt",
    },
    "bart-base": {
        "config": "https://huggingface.co/facebook/bart-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/facebook/bart-base/resolve/main/vocab.json",
        "merge": "https://huggingface.co/facebook/bart-base/resolve/main/merges.txt",
        "weight": "https://huggingface.co/facebook/bart-base/resolve/main/pytorch_model.bin",
    },
    "bart-large": {
        "config": "https://huggingface.co/facebook/bart-large/resolve/main/config.json",
        "vocab": "https://huggingface.co/facebook/bart-large/resolve/main/vocab.json",
        "merge": "https://huggingface.co/facebook/bart-large/resolve/main/merges.txt",
        "weight": "https://huggingface.co/facebook/bart-large/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.bart.modeling
import unitorch.cli.models.bart.processing
from unitorch.cli.models.bart.modeling import BartForGeneration
from unitorch.cli.models.bart.processing import BartProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
import torch.nn as nn
import pyarrow as pa
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.models import GenericOnnxModel
from unitorch.cli import (
    add_default_section_for_init,
    add_default_section_for_function,
    register_process,
)
from unitorch.cli import WriterMixin, WriterOutputs
from unitorch.cli.models import TensorsOutputs, TensorsTargets, ACT2FN


def onnx_model_decorator(cls):
    class OnnxModel(GenericOnnxModel):
        def __init__(self, *args, **kwargs):
            super().__init__()

    return OnnxModel



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_mt5_infos = {
    "default-mt5": {
        "config": "https://huggingface.co/google/mt5-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/mt5-base/resolve/main/spiece.model",
    },
    "mt5-base": {
        "config": "https://huggingface.co/google/mt5-base/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/mt5-base/resolve/main/spiece.model",
        "weight": "https://huggingface.co/google/mt5-base/resolve/main/pytorch_model.bin",
    },
    "mt5-small": {
        "config": "https://huggingface.co/google/mt5-small/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/mt5-small/resolve/main/spiece.model",
        "weight": "https://huggingface.co/google/mt5-small/resolve/main/pytorch_model.bin",
    },
    "mt5-large": {
        "config": "https://huggingface.co/google/mt5-large/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/mt5-large/resolve/main/spiece.model",
        "weight": "https://huggingface.co/google/mt5-large/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.mt5.modeling
import unitorch.cli.models.mt5.processing
from unitorch.cli.models.mt5.modeling import MT5ForGeneration
from unitorch.cli.models.mt5.processing import MT5Processor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import io
import torch
import gc
import gradio as gr
from typing import List, Tuple
from PIL import Image, ImageDraw
from unitorch.cli import CoreConfigureParser, GenericWebUI
from unitorch.cli import register_webui
from unitorch.cli.models.detr import pretrained_detr_infos
from unitorch.cli.pipelines.detr import DetrForDetectionPipeline
from unitorch.cli.webuis import (
    matched_pretrained_names,
    create_element,
    create_accordion,
    create_row,
    create_column,
    create_group,
    create_tab,
    create_tabs,
    create_blocks,
    create_pretrain_layout,
    create_freeu_layout,
)
from unitorch.cli.webuis import SimpleWebUI


@register_webui("core/webui/detr")
class DetrWebUI(SimpleWebUI):
    pretrained_names = list(pretrained_detr_infos.keys())
    supported_pretrained_names = matched_pretrained_names(
        pretrained_names,
        "^detr-",
    )

    def __init__(self, config: CoreConfigureParser):
        self._config = config
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        if len(self.supported_pretrained_names) == 0:
            raise ValueError("No supported pretrained models found.")
        self._name = self.supported_pretrained_names[0]

        # create elements
        pretrain_layout_group = create_pretrain_layout(
            self.supported_pretrained_names, self._name
        )
        name, status, start, stop, pretrain_layout = (
            pretrain_layout_group.name,
            pretrain_layout_group.status,
            pretrain_layout_group.start,
            pretrain_layout_group.stop,
            pretrain_layout_group.layout,
        )

        input_image = create_element("image", "Input Image")
        mask_threshold = create_element(
            "slider", "Mask Threshold", default=0.5, min_value=0, max_value=1, step=0.01
        )
        detect = create_element("button", "Detect")
        output_image = create_element("image", "Output Image")

        # create blocks
        left = create_column(input_image, mask_threshold, detect)
        right = create_column(output_image)
        iface = create_blocks(pretrain_layout, create_row(left, right))

        # create events
        iface.__enter__()

        start.click(self.start, inputs=[name], outputs=[status])
        stop.click(self.stop, outputs=[status])
        detect.click(
            self.serve, inputs=[input_image, mask_threshold], outputs=[output_image]
        )

        iface.load(
            fn=lambda: [gr.update(value=self._name), gr.update(value=self._status)],
            outputs=[name, status],
        )

        iface.__exit__()

        super().__init__(config, iname="DETR", iface=iface)

    def start(self, pretrained_name, **kwargs):
        if self._status == "Running":
            self.stop()
        self._name = pretrained_name
        self._pipe = DetrForDetectionPipeline.from_core_configure(
            self._config,
            pretrained_name=pretrained_name,
        )
        self._status = "Running"
        return self._status

    def stop(self, **kwargs):
        self._pipe.to("cpu")
        del self._pipe
        gc.collect()
        torch.cuda.empty_cache()
        self._pipe = None if not hasattr(self, "_pipe") else self._pipe
        self._status = "Stopped" if self._pipe is None else "Running"
        return self._status

    def serve(
        self,
        image: Image.Image,
        mask_threshold: float = 0.5,
    ):
        assert self._pipe is not None
        result = self._pipe(
            image,
            threshold=mask_threshold,
        )
        return result



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import json
import logging
import torch
from PIL import Image
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from diffusers.utils import numpy_to_pil
from diffusers.models import SD3ControlNetModel, SD3MultiControlNetModel
from diffusers.pipelines import (
    StableDiffusion3Pipeline,
    StableDiffusion3Img2ImgPipeline,
    StableDiffusion3ControlNetPipeline,
)
from unitorch import is_xformers_available
from unitorch.utils import is_remote_url
from unitorch.models.diffusers import GenericStable3Model
from unitorch.models.diffusers import Stable3Processor

from unitorch.utils import pop_value, nested_dict_value
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
)
from unitorch.cli.models.diffusers import (
    pretrained_stable_infos,
    pretrained_stable_extensions_infos,
    load_weight,
)
from unitorch.cli.pipelines import Schedulers


class Stable3ForText2ImageGenerationPipeline(GenericStable3Model):
    def __init__(
        self,
        config_path: str,
        text_config_path: str,
        text2_config_path: str,
        text3_config_path: str,
        vae_config_path: str,
        scheduler_config_path: str,
        vocab_path: str,
        merge_path: str,
        vocab2_path: str,
        merge2_path: str,
        vocab3_path: str,
        quant_config_path: Optional[str] = None,
        max_seq_length: Optional[int] = 77,
        max_seq_length2: Optional[int] = 256,
        pad_token: Optional[str] = "<|endoftext|>",
        pad_token2: Optional[str] = "!",
        weight_path: Optional[Union[str, List[str]]] = None,
        state_dict: Optional[Dict[str, Any]] = None,
        device: Optional[Union[str, int]] = "cpu",
        enable_cpu_offload: Optional[bool] = False,
        enable_xformers: Optional[bool] = False,
    ):
        super().__init__(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            text3_config_path=text3_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            quant_config_path=quant_config_path,
        )
        self.processor = Stable3Processor(
            vocab_path=vocab_path,
            merge_path=merge_path,
            vocab2_path=vocab2_path,
            merge2_path=merge2_path,
            vocab3_path=vocab3_path,
            vae_config_path=vae_config_path,
            max_seq_length=max_seq_length,
            max_seq_length2=max_seq_length2,
            pad_token=pad_token,
            pad_token2=pad_token2,
        )
        self._device = "cpu" if device == "cpu" else int(device)

        self.from_pretrained(weight_path, state_dict=state_dict)
        self.eval()
        self.to(device=self._device)

        self._enable_cpu_offload = enable_cpu_offload
        self._enable_xformers = enable_xformers

    @classmethod
    @add_default_section_for_init("core/pipeline/stable_3/text2image")
    def from_core_configure(
        cls,
        config,
        pretrained_name: Optional[str] = "stable-v3-base",
        config_path: Optional[str] = None,
        text_config_path: Optional[str] = None,
        text2_config_path: Optional[str] = None,
        text3_config_path: Optional[str] = None,
        vae_config_path: Optional[str] = None,
        scheduler_config_path: Optional[str] = None,
        vocab_path: Optional[str] = None,
        merge_path: Optional[str] = None,
        vocab2_path: Optional[str] = None,
        merge2_path: Optional[str] = None,
        vocab3_path: Optional[str] = None,
        quant_config_path: Optional[str] = None,
        pretrained_weight_path: Optional[str] = None,
        device: Optional[str] = "cpu",
        **kwargs,
    ):
        config.set_default_section("core/pipeline/stable_3/text2image")
        pretrained_name = config.getoption("pretrained_name", pretrained_name)
        pretrained_infos = nested_dict_value(pretrained_stable_infos, pretrained_name)

        config_path = config.getoption("config_path", config_path)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_infos, "transformer", "config"),
        )
        config_path = cached_path(config_path)

        text_config_path = config.getoption("text_config_path", text_config_path)
        text_config_path = pop_value(
            text_config_path,
            nested_dict_value(pretrained_infos, "text", "config"),
        )
        text_config_path = cached_path(text_config_path)

        text2_config_path = config.getoption("text2_config_path", text2_config_path)
        text2_config_path = pop_value(
            text2_config_path,
            nested_dict_value(pretrained_infos, "text2", "config"),
        )
        text2_config_path = cached_path(text2_config_path)

        text3_config_path = config.getoption("text3_config_path", text3_config_path)
        text3_config_path = pop_value(
            text3_config_path,
            nested_dict_value(pretrained_infos, "text3", "config"),
        )
        text3_config_path = cached_path(text3_config_path)

        vae_config_path = config.getoption("vae_config_path", vae_config_path)
        vae_config_path = pop_value(
            vae_config_path,
            nested_dict_value(pretrained_infos, "vae", "config"),
        )
        vae_config_path = cached_path(vae_config_path)

        scheduler_config_path = config.getoption(
            "scheduler_config_path", scheduler_config_path
        )
        scheduler_config_path = pop_value(
            scheduler_config_path,
            nested_dict_value(pretrained_infos, "scheduler"),
        )
        scheduler_config_path = cached_path(scheduler_config_path)

        vocab_path = config.getoption("vocab_path", vocab_path)
        vocab_path = pop_value(
            vocab_path,
            nested_dict_value(pretrained_infos, "text", "vocab"),
        )
        vocab_path = cached_path(vocab_path)

        merge_path = config.getoption("merge_path", merge_path)
        merge_path = pop_value(
            merge_path,
            nested_dict_value(pretrained_infos, "text", "merge"),
        )
        merge_path = cached_path(merge_path)

        vocab2_path = config.getoption("vocab2_path", vocab2_path)
        vocab2_path = pop_value(
            vocab2_path,
            nested_dict_value(pretrained_infos, "text2", "vocab"),
        )
        vocab2_path = cached_path(vocab2_path)

        merge2_path = config.getoption("merge2_path", merge2_path)
        merge2_path = pop_value(
            merge2_path,
            nested_dict_value(pretrained_infos, "text2", "merge"),
        )
        merge2_path = cached_path(merge2_path)

        vocab3_path = config.getoption("vocab3_path", vocab3_path)
        vocab3_path = pop_value(
            vocab3_path,
            nested_dict_value(pretrained_infos, "text3", "vocab"),
        )
        vocab3_path = cached_path(vocab3_path)

        quant_config_path = config.getoption("quant_config_path", quant_config_path)
        if quant_config_path is not None:
            quant_config_path = cached_path(quant_config_path)

        max_seq_length = config.getoption("max_seq_length", 77)
        max_seq_length2 = config.getoption("max_seq_length2", 256)
        pad_token = config.getoption("pad_token", "<|endoftext|>")
        pad_token2 = config.getoption("pad_token2", "!")
        weight_path = config.getoption("pretrained_weight_path", pretrained_weight_path)
        device = config.getoption("device", device)
        enable_cpu_offload = config.getoption("enable_cpu_offload", True)
        enable_xformers = config.getoption("enable_xformers", True)

        state_dict = None
        if weight_path is None and pretrained_infos is not None:
            state_dict = [
                load_weight(
                    nested_dict_value(pretrained_infos, "transformer", "weight"),
                    prefix_keys={"": "transformer."},
                ),
                load_weight(
                    nested_dict_value(pretrained_infos, "text", "weight"),
                    prefix_keys={"": "text."},
                ),
                load_weight(
                    nested_dict_value(pretrained_infos, "text2", "weight"),
                    prefix_keys={"": "text2."},
                ),
                load_weight(
                    nested_dict_value(pretrained_infos, "text3", "weight"),
                    prefix_keys={"": "text3."},
                ),
                load_weight(
                    nested_dict_value(pretrained_infos, "vae", "weight"),
                    prefix_keys={"": "vae."},
                ),
            ]

        inst = cls(
            config_path=config_path,
            text_config_path=text_config_path,
            text2_config_path=text2_config_path,
            text3_config_path=text3_config_path,
            vae_config_path=vae_config_path,
            scheduler_config_path=scheduler_config_path,
            vocab_path=vocab_path,
            merge_path=merge_path,
            vocab2_path=vocab2_path,
            merge2_path=merge2_path,
            vocab3_path=vocab3_path,
            quant_config_path=quant_config_path,
            max_seq_length=max_seq_length,
            max_seq_length2=max_seq_length2,
            pad_token=pad_token,
            pad_token2=pad_token2,
            weight_path=weight_path,
            state_dict=state_dict,
            device=device,
            enable_cpu_offload=enable_cpu_offload,
            enable_xformers=enable_xformers,
        )
        return inst

    @torch.no_grad()
    @add_default_section_for_function("core/pipeline/stable_3/text2image")
    def __call__(
        self,
        text: str,
        neg_text: Optional[str] = "",
        height: Optional[int] = 1024,
        width: Optional[int] = 1024,
        guidance_scale: Optional[float] = 7.5,
        num_timesteps: Optional[int] = 50,
        seed: Optional[int] = 1123,
        scheduler: Optional[str] = None,
        freeu_params: Optional[Tuple[float, float, float, float]] = (
            0.9,
            0.2,
            1.2,
            1.4,
        ),
        controlnet_checkpoints: Optional[List[str]] = None,
        controlnet_images: Optional[List[Image.Image]] = None,
        controlnet_guidance_scales: Optional[List[float]] = None,
        lora_checkpoints: Optional[Union[str, List[str]]] = None,
        lora_weights: Optional[Union[float, List[float]]] = 1.0,
        lora_alphas: Optional[Union[float, List[float]]] = 32,
        lora_urls: Optional[Union[str, List[str]]] = None,
        lora_files: Optional[Union[str, List[str]]] = None,
    ):
        text_inputs = self.processor.text2image_inputs(
            text,
            negative_prompt=neg_text,
        )
        assert scheduler is None or scheduler in Schedulers
        if scheduler is not None:
            self.scheduler = Schedulers.get(scheduler).from_config(
                self.scheduler.config
            )
        self.scheduler.set_timesteps(num_inference_steps=num_timesteps)

        if any(ckpt is not None for ckpt in controlnet_checkpoints) and any(
            img is not None for img in controlnet_images
        ):
            controlnets, conditioning_scales, conditioning_images = [], [], []
            for checkpoint, conditioning_scale, conditioning_image in zip(
                controlnet_checkpoints, controlnet_guidance_scales, controlnet_images
            ):
                if checkpoint is None or conditioning_image is None:
                    continue
                controlnet_config_path = cached_path(
                    nested_dict_value(
                        pretrained_stable_extensions_infos,
                        checkpoint,
                        "controlnet",
                        "config",
                    )
                )
                controlnet_config_dict = json.load(open(controlnet_config_path))
                controlnet = SD3ControlNetModel.from_config(controlnet_config_dict)
                controlnet.load_state_dict(
                    load_weight(
                        nested_dict_value(
                            pretrained_stable_extensions_infos,
                            checkpoint,
                            "controlnet",
                            "weight",
                        )
                    ),
                )
                controlnet.to(device=self._device)
                logging.info(f"Loading controlnet from {checkpoint}")
                controlnets.append(controlnet)
                controlnets = SD3MultiControlNetModel(controlnets)
                conditioning_scales.append(conditioning_scale)
                conditioning_images.append(conditioning_image.resize((width, height)))
            self.pipeline = StableDiffusion3ControlNetPipeline(
                vae=self.vae,
                text_encoder=self.text,
                text_encoder_2=self.text2,
                text_encoder_3=self.text3,
                transformer=self.transformer,
                controlnet=controlnets,
                scheduler=self.scheduler,
                tokenizer=None,
                tokenizer_2=None,
                tokenizer_3=None,
            )
            controlnets_inputs = self.processor.controlnets_inputs(conditioning_images)
            enable_controlnet = True
            inputs = {
                **text_inputs,
                **{"condition_pixel_values": controlnets_inputs.pixel_values},
            }
        else:
            self.pipeline = StableDiffusion3Pipeline(
                vae=self.vae,
                text_encoder=self.text,
                text_encoder_2=self.text2,
                text_encoder_3=self.text3,
                transformer=self.transformer,
                scheduler=self.scheduler,
                tokenizer=None,
                tokenizer_2=None,
                tokenizer_3=None,
            )
            enable_controlnet = False
            inputs = text_inputs
        self.pipeline.set_progress_bar_config(disable=True)
        # self.pipeline.enable_freeu(*freeu_params)
        self.seed = seed

        inputs = {k: v.unsqueeze(0) if v is not None else v for k, v in inputs.items()}
        inputs = {
            k: v.to(device=self._device) if v is not None else v
            for k, v in inputs.items()
        }
        if isinstance(lora_checkpoints, str):
            lora_checkpoints = [lora_checkpoints]
        if isinstance(lora_weights, float):
            lora_weights = [lora_weights]
        if isinstance(lora_alphas, float):
            lora_alphas = [lora_alphas]
        if isinstance(lora_urls, str):
            lora_urls = [lora_urls]
        if isinstance(lora_files, str):
            lora_files = [lora_files]

        if lora_checkpoints is not None:
            _lora_checkpoints = list(
                zip(
                    *[
                        (ckpt, weight, alpha)
                        for ckpt, weight, alpha in zip(
                            lora_checkpoints, lora_weights, lora_alphas
                        )
                        if ckpt is not None
                    ]
                )
            )
            if len(_lora_checkpoints) == 0:
                lora_checkpoints = None
            else:
                lora_checkpoints, lora_weights, lora_alphas = _lora_checkpoints
                lora_checkpoints = [
                    nested_dict_value(pretrained_stable_extensions_infos, ckpt)
                    for ckpt in lora_checkpoints
                ]
        if lora_urls is not None:
            _lora_urls = list(
                zip(
                    *[
                        (url, weight, alpha)
                        for url, weight, alpha in zip(
                            lora_urls, lora_weights, lora_alphas
                        )
                        if is_remote_url(url)
                    ]
                )
            )
            if len(_lora_urls) == 0:
                lora_urls = None
            else:
                lora_urls, lora_weights, lora_alphas = _lora_urls
        if lora_files is not None:
            _lora_files = list(
                zip(
                    *[
                        (file, weight, alpha)
                        for file, weight, alpha in zip(
                            lora_files, lora_weights, lora_alphas
                        )
                        if file is not None
                    ]
                )
            )
            if len(_lora_files) == 0:
                lora_files = None
            else:
                lora_files, lora_weights, lora_alphas = _lora_files

        lora_files = pop_value(
            lora_checkpoints,
            lora_urls,
            lora_files,
            check_none=False,
        )
        if lora_files is not None:
            self.load_lora_weights(
                lora_files, lora_weights=lora_weights, lora_alphas=lora_alphas
            )

        prompt_embeds_results = self.get_prompt_outputs(
            inputs["input_ids"],
            input2_ids=inputs["input2_ids"],
            input3_ids=inputs["input3_ids"],
            negative_input_ids=inputs["negative_input_ids"],
            negative_input2_ids=inputs["negative_input2_ids"],
            negative_input3_ids=inputs["negative_input3_ids"],
        )
        prompt_embeds = prompt_embeds_results.prompt_embeds
        negative_prompt_embeds = prompt_embeds_results.negative_prompt_embeds
        pooled_prompt_embeds = prompt_embeds_results.pooled_prompt_embeds
        negative_pooled_prompt_embeds = (
            prompt_embeds_results.negative_pooled_prompt_embeds
        )

        if self._enable_cpu_offload and self._device != "cpu":
            self.pipeline.enable_model_cpu_offload(self._device)
        else:
            self.to(device=self._device)

        if self._enable_xformers and self._device != "cpu":
            assert is_xformers_available(), "Please install xformers first."
            self.pipeline.enable_xformers_memory_efficient_attention()

        if enable_controlnet:
            outputs = self.pipeline(
                prompt_embeds=prompt_embeds,
                negative_prompt_embeds=negative_prompt_embeds,
                pooled_prompt_embeds=pooled_prompt_embeds,
                negative_pooled_prompt_embeds=negative_pooled_prompt_embeds,
                generator=torch.Generator(device=self.pipeline.device).manual_seed(
                    self.seed
                ),
                height=height,
                width=width,
                control_image=list(inputs["condition_pixel_values"].transpose(0, 1)),
                guidance_scale=guidance_scale,
                controlnet_conditioning_scale=conditioning_scales,
                output_type="np.array",
            )
        else:
            outputs = self.pipeline(
                prompt_embeds=prompt_embeds,
                negative_prompt_embeds=negative_prompt_embeds,
                pooled_prompt_embeds=pooled_prompt_embeds,
                negative_pooled_prompt_embeds=negative_pooled_prompt_embeds,
                generator=torch.Generator(device=self.pipeline.device).manual_seed(
                    self.seed
                ),
                height=height,
                width=width,
                guidance_scale=guidance_scale,
                output_type="np.array",
            )
        self.unload_lora_weights()

        images = torch.from_numpy(outputs.images)
        images = numpy_to_pil(images.cpu().numpy())
        return images[0]



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import ast
import torch
import datasets
import torch.distributed as dist
from copy import deepcopy
from datasets import Dataset
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.datasets.hf import HFDatasets, HFIterableDatasets
from unitorch.cli import CoreConfigureParser
from unitorch.cli import cached_path, registered_process, register_dataset
from unitorch.cli import add_default_section_for_init, add_default_section_for_function
from unitorch.cli import init_registered_process
from unitorch.cli.models import (
    ModelInputs,
    ModelTargets,
    CombineTensorsInputs,
    CombineTensorsTargets,
)


class ASTFunction:
    def __init__(self, func: str):
        for name in registered_process.keys():
            _name = f"{name}("
            func = func.replace(_name, _name.replace("/", "_"))

        registered_process_mapping = {
            k.replace("/", "_"): k for k, v in registered_process.items()
        }
        self.func = func
        self.__ast_func__ = ast.parse(func, "", mode="eval")
        self.__ast_keys__ = []
        self.__ast_process__ = []
        for node in ast.walk(self.__ast_func__):
            if isinstance(node, ast.Name):
                if node.id in registered_process_mapping:
                    self.__ast_process__.append(node.id)
                else:
                    self.__ast_keys__.append(node.id)
        self.__ast_func__ = compile(self.__ast_func__, "", "eval")

    def process(self, row: Dict):
        for key in self.__ast_keys__:
            if key not in row:
                continue
            locals()[key] = row.get(key, None)

        return eval(deepcopy(self.__ast_func__))


class ASTHFDatasets(HFDatasets):
    def __init__(
        self,
        dataset: Dataset,
        process_functions: List[ASTFunction],
    ):
        super().__init__(dataset=dataset)
        self.process_functions = deepcopy(process_functions)

    def __getitem__(self, idx):
        model_inputs_list, model_targets_list = [], []
        row = self.dataset[idx]
        for pfunc in self.process_functions:
            results = pfunc.process(row)
            if isinstance(results, ModelInputs) or isinstance(results, ModelTargets):
                results = [results]

            for result in results:
                if isinstance(result, ModelInputs):
                    is_new = True
                    for _inputs in model_inputs_list:
                        if type(_inputs) == type(result):
                            _inputs.add(result)
                            is_new = False
                            break

                    if is_new:
                        model_inputs_list.append(result)

                if isinstance(result, ModelTargets):
                    is_new = True
                    for _targets in model_targets_list:
                        if type(_targets) == type(result):
                            _targets.add(result)
                            is_new = False
                            break

                    if is_new:
                        model_targets_list.append(result)
        if len(model_inputs_list) == 0:
            model_inputs_list = CombineTensorsInputs()
        elif len(model_inputs_list) == 1:
            model_inputs_list = model_inputs_list[0]

        if len(model_targets_list) == 0:
            model_targets_list = CombineTensorsTargets()
        elif len(model_targets_list) == 1:
            model_targets_list = model_targets_list[0]

        return model_inputs_list, model_targets_list

    def __len__(
        self,
    ):
        return len(self.dataset)


class ASTHFIterableDatasets(HFIterableDatasets):
    def __init__(
        self,
        dataset: Dataset,
        process_functions: List[ASTFunction],
        enable_ddp_partition: Optional[bool] = True,
    ):
        super().__init__(dataset=dataset)
        self.process_functions = deepcopy(process_functions)
        self.dataset.shuffle(10000)
        if enable_ddp_partition and dist.is_initialized():
            self.global_rank = dist.get_rank()
            self.world_size = dist.get_world_size()
        else:
            self.global_rank = 0
            self.world_size = 1

    def __iter__(self):
        worker_info = torch.utils.data.get_worker_info()
        mod = self.world_size
        shift = self.global_rank
        if worker_info:
            mod *= worker_info.num_workers
            shift = self.global_rank * worker_info.num_workers + worker_info.id
        for i, row in enumerate(self.dataset):
            if (i + shift) % mod != 0:
                continue
            inputs, targets = None, None
            for pfunc in self.process_functions:
                results = pfunc.process(row)

                if isinstance(results, ModelInputs) or isinstance(
                    results, ModelTargets
                ):
                    results = [results]

                for result in results:
                    if isinstance(result, ModelInputs):
                        if inputs is None:
                            inputs = result
                        else:
                            inputs.add(result)

                    if isinstance(result, ModelTargets):
                        if targets is None:
                            targets = result
                        else:
                            targets.add(result)
            if inputs is None:
                inputs = CombineTensorsInputs()

            if targets is None:
                targets = CombineTensorsTargets()

            yield inputs, targets

    def set_skip_step(self, step):
        self.__dataset__ = self.__dataset__.skip(step * self.world_size)


@register_dataset("core/dataset/ast")
class ASTDatasets:
    """Class for managing AST datasets."""

    splits = ["train", "dev", "test"]
    templates = ["csv", "json", "parquet", "hub"]
    __ASTDatasets__ = dict()

    def __init__(self, configure: CoreConfigureParser):
        """
        Initialize ASTDatasets.

        Args:
            configure (CoreConfigureParser): The configuration parser for AST datasets.
        """
        self.config = configure

    def __getdataset__(self, split):
        """
        Get the dataset for the specified split.

        Args:
            split (str): The split to get the dataset for.

        Returns:
            dataset: The dataset for the specified split.
        """
        config = self.config

        registered_process_mapping = {
            k.replace("/", "_"): k for k, v in registered_process.items()
        }

        config.set_default_section(f"core/dataset/ast")
        _iterable = config.getoption("iterable", False)
        _template = config.getoption("template", "csv")
        _data_name = config.getoption("data_name", None)
        _config_name = config.getoption("config_name", None)
        _data_dir = config.getoption("data_dir", None)
        _data_files = config.getoption("data_files", None)
        _names = config.getoption("names", None)
        _features = config.getoption("features", None)
        _sep = config.getoption("sep", "\t")
        _quoting = config.getoption("quoting", 3)
        _escapechar = config.getoption("escapechar", None)
        _field = config.getoption("field", None)
        _process_functions = config.getoption("preprocess_functions", None)
        _enable_ddp_partition = config.getoption("enable_ddp_partition", True)

        _HFDatasets = HFIterableDatasets if _iterable else HFDatasets
        _ASTDatasets = ASTHFIterableDatasets if _iterable else ASTHFDatasets

        config.set_default_section(f"core/dataset/ast/{split}")

        template = config.getoption("template", _template)
        if config.getoption("data_name", _data_name) is not None:
            template = "hub"

        assert template in self.templates

        new_split = "validation" if split == "dev" else split
        new_split = config.getoption("split", new_split)

        # get dataset
        dataset = None
        if template == "csv":
            data_dir = config.getoption("data_dir", _data_dir)
            data_files = config.getoption("data_files", _data_files)
            names = config.getoption("names", _names)
            sep = config.getoption("sep", _sep)
            quoting = config.getoption("quoting", _quoting)
            escapechar = config.getoption("escapechar", _escapechar)
            dataset = _HFDatasets.from_csv(
                data_dir=data_dir,
                data_files=data_files,
                names=names,
                sep=sep,
                quoting=quoting,
                escapechar=escapechar,
                split=new_split,
            )

        if template == "json":
            data_dir = config.getoption("data_dir", _data_dir)
            data_files = config.getoption("data_files", _data_files)
            field = config.getoption("field", _field)

            dataset = _HFDatasets.from_json(
                data_dir=data_dir,
                data_files=data_files,
                field=field,
                split=new_split,
            )

        if template == "parquet":
            data_dir = config.getoption("data_dir", _data_dir)
            data_files = config.getoption("data_files", _data_files)
            features = config.getoption("features", _features)
            if isinstance(features, str):
                features = eval(features)
            dataset = _HFDatasets.from_parquet(
                data_dir=data_dir,
                data_files=data_files,
                split=new_split,
                features=features,
            )

        if template == "hub":
            data_name = config.getoption("data_name", _data_name)
            config_name = config.getoption("config_name", _config_name)
            data_dir = config.getoption("data_dir", _data_dir)
            data_files = config.getoption("data_files", _data_files)
            data_name = (
                cached_path(data_name) if data_name.endswith(".py") else data_name
            )
            dataset = _HFDatasets.from_hub(
                data_name=data_name,
                config_name=config_name,
                data_dir=data_dir,
                data_files=data_files,
                split=new_split,
            )

        assert dataset is not None

        # get process functions
        process_functions = config.getoption("preprocess_functions", _process_functions)
        if process_functions is None:
            process_functions = []
        else:
            process_functions = [ASTFunction(func) for func in process_functions]

        for pfunc in process_functions:
            for name in pfunc.__ast_process__:
                globals()[name] = init_registered_process(
                    registered_process_mapping[name],
                    config,
                )

        enable_ddp_partition = config.getoption(
            "enable_ddp_partition", _enable_ddp_partition
        )

        if isinstance(_ASTDatasets, HFIterableDatasets):
            self.__ASTDatasets__[split] = _ASTDatasets(
                dataset=dataset.dataset,
                process_functions=process_functions,
                enable_ddp_partition=enable_ddp_partition,
            )
        else:
            self.__ASTDatasets__[split] = _ASTDatasets(
                dataset=dataset.dataset,
                process_functions=process_functions,
            )

        return self.__ASTDatasets__.get(split)

    @classmethod
    @add_default_section_for_init("core/dataset/ast")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of ASTDatasets from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            ASTDatasets: An instance of ASTDatasets.
        """
        return cls(configure=config)

    def get(self, split: Optional[str] = "train"):
        """
        Get the dataset for the specified split.

        Args:
            split (str, optional): The split to get the dataset for. Defaults to "train".

        Returns:
            dataset: The dataset for the specified split.
        """
        return self.__getdataset__(split)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from functools import partial

import torch
from transformers import DebertaV2Tokenizer
from unitorch.utils import pop_value, truncate_sequence_pair
from unitorch.models import GenericOutputs, HfTextClassificationProcessor


def get_deberta_v2_tokenizer(vocab_path):
    assert os.path.exists(vocab_path)
    tokenizer = DebertaV2Tokenizer(vocab_path)
    return tokenizer


class DebertaV2Processor(HfTextClassificationProcessor):
    """
    Processor for DeBERTa-v2 based text classification models.
    """

    def __init__(
        self,
        vocab_path: str,
        max_seq_length: Optional[int] = 128,
        source_type_id: Optional[int] = 0,
        target_type_id: Optional[int] = 1,
    ):
        """
        Initializes the DebertaV2Processor.

        Args:
            vocab_path (str): Path to the vocabulary file.
            max_seq_length (Optional[int]): Maximum sequence length. Defaults to 128.
            source_type_id (Optional[int]): Source type ID for token types. Defaults to 0.
            target_type_id (Optional[int]): Target type ID for token types. Defaults to 1.
        """
        tokenizer = get_deberta_v2_tokenizer(
            vocab_path,
        )
        super().__init__(
            tokenizer=tokenizer,
            max_seq_length=max_seq_length,
            source_type_id=source_type_id,
            target_type_id=target_type_id,
            position_start_id=tokenizer.pad_token_id + 1,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

# pretrained infos
pretrained_pegasus_infos = {
    "default-pegasus": {
        "config": "https://huggingface.co/google/pegasus-cnn_dailymail/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/pegasus-cnn_dailymail/resolve/main/spiece.model",
    },
    "pegasus-cnn_dailymail": {
        "config": "https://huggingface.co/google/pegasus-cnn_dailymail/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/pegasus-cnn_dailymail/resolve/main/spiece.model",
        "weight": "https://huggingface.co/google/pegasus-cnn_dailymail/resolve/main/pytorch_model.bin",
    },
    "pegasus-xsum": {
        "config": "https://huggingface.co/google/pegasus-xsum/resolve/main/config.json",
        "vocab": "https://huggingface.co/google/pegasus-xsum/resolve/main/spiece.model",
        "weight": "https://huggingface.co/google/pegasus-xsum/resolve/main/pytorch_model.bin",
    },
}

import unitorch.cli.models.pegasus.modeling
import unitorch.cli.models.pegasus.processing
from unitorch.cli.models.pegasus.modeling import PegasusForGeneration
from unitorch.cli.models.pegasus.processing import PegasusProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import torch
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from torch.cuda.amp import autocast
from unitorch.utils import pop_value, nested_dict_value
from unitorch.models.detr import DetrForDetection as _DetrForDetection
from unitorch.cli import (
    cached_path,
    add_default_section_for_init,
    add_default_section_for_function,
    register_model,
)
from unitorch.cli.models import (
    detection_model_decorator,
    segmentation_model_decorator,
    DetectionOutputs,
    SegmentationOutputs,
    LossOutputs,
)
from unitorch.cli.models.detr import pretrained_detr_infos


@register_model("core/model/detection/detr", detection_model_decorator)
class DetrForDetection(_DetrForDetection):
    def __init__(
        self,
        config_path: str,
        num_classes: Optional[int] = None,
    ):
        super().__init__(
            config_path=config_path,
            num_classes=num_classes,
        )

    @classmethod
    @add_default_section_for_init("core/model/detection/detr")
    def from_core_configure(cls, config, **kwargs):
        config.set_default_section("core/model/detection/detr")
        pretrained_name = config.getoption("pretrained_name", "default-detr")
        config_path = config.getoption("config_path", None)
        config_path = pop_value(
            config_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "config"),
        )

        config_path = cached_path(config_path)
        num_classes = config.getoption("num_classes", None)

        inst = cls(
            config_path=config_path,
            num_classes=num_classes,
        )
        pretrained_weight_path = config.getoption("pretrained_weight_path", None)
        weight_path = pop_value(
            pretrained_weight_path,
            nested_dict_value(pretrained_detr_infos, pretrained_name, "weight"),
            check_none=False,
        )
        if weight_path is not None:
            inst.from_pretrained(weight_path)

        return inst

    @autocast()
    def forward(self, images, bboxes, classes):
        outputs = super().forward(
            images=images,
            bboxes=bboxes,
            classes=classes,
        )
        return LossOutputs(loss=outputs)

    @add_default_section_for_function("core/model/detection/detr")
    def detect(
        self,
        images,
        norm_bboxes: bool = False,
    ):
        outputs = super().detect(
            images=images,
            norm_bboxes=norm_bboxes,
        )
        return DetectionOutputs(
            bboxes=outputs.bboxes,
            scores=outputs.scores,
            classes=outputs.classes,
        )



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from unitorch.models.llava.modeling import (
    LlavaMistralClipForClassification,
    LlavaMistralClipForGeneration,
)
from unitorch.models.llava.processing import LlavaMistralClipProcessor



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
import gradio as gr
from unitorch.models import GenericOutputs


def create_element(
    dtype,
    label,
    default=None,
    values=[],
    min_value=None,
    max_value=None,
    step=None,
    scale=None,
    multiselect=None,
    info=None,
    interactive=None,
    variant=None,
    lines=1,
    placeholder=None,
    show_label=True,
    elem_id=None,
    elem_classes=None,
):
    if dtype == "text":
        return gr.Textbox(
            value=default,
            label=label,
            scale=scale,
            info=info,
            interactive=interactive,
            lines=lines,
            placeholder=placeholder,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "number":
        return gr.Number(
            value=default,
            label=label,
            scale=scale,
            info=info,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "slider":
        return gr.Slider(
            value=default,
            label=label,
            minimum=min_value,
            maximum=max_value,
            step=step,
            scale=scale,
            info=info,
            interactive=interactive,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "checkbox":
        return gr.Checkbox(
            value=default,
            label=label,
            scale=scale,
            info=info,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "checkboxgroup":
        return gr.CheckboxGroup(
            value=default,
            label=label,
            choices=values,
            scale=scale,
            info=info,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "radio":
        return gr.Radio(
            value=default,
            label=label,
            choices=values,
            scale=scale,
            info=info,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "dropdown":
        return gr.Dropdown(
            value=default,
            label=label,
            choices=values,
            scale=scale,
            multiselect=multiselect,
            info=info,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "dataframe":
        return gr.Dataframe(
            value=default,
            label=label,
            scale=scale,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
            datatype="markdown",
            wrap=True,
            height=1000,
        )

    if dtype == "image":
        return gr.Image(
            type="pil",
            value=default,
            label=label,
            scale=scale,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "image_editor":
        return gr.ImageEditor(
            type="pil",
            value=default,
            label=label,
            scale=scale,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
            eraser=gr.Eraser(),
            brush=gr.Brush(),
            canvas_size=(1024, 1024),
        )

    if dtype == "audio":
        return gr.Audio(
            label=label,
            scale=scale,
            info=info,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "video":
        return gr.Video(
            label=label,
            scale=scale,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "file":
        return gr.File(
            label=label,
            scale=scale,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "gallery":
        return gr.Gallery(
            label=label,
            scale=scale,
            info=info,
            interactive=interactive,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "button":
        interactive = True if interactive is None else interactive
        variant = "primary" if variant is None else variant
        return gr.Button(
            value=label,
            scale=scale,
            interactive=interactive,
            variant=variant,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    if dtype == "markdown":
        return gr.Markdown(
            value=label,
            show_label=show_label,
            elem_id=elem_id,
            elem_classes=elem_classes,
        )

    raise ValueError(f"Unknown element type: {dtype}")


def create_accordion(*elements, name=None, open=False):
    accordion = gr.Accordion(label=name, open=open)
    for element in elements:
        accordion.add_child(element)
    return accordion


def create_row(*elements, variant="panel"):
    row = gr.Row(variant=variant)
    for element in elements:
        row.add_child(element)
    return row


def create_column(*elements, variant="panel", scale=1):
    col = gr.Column(
        variant=variant,
        scale=scale,
    )
    for element in elements:
        col.add_child(element)
    return col


def create_group(*elements):
    group = gr.Group()
    for element in elements:
        group.add_child(element)
    return group


def create_tab(*elements, name=None):
    tab = gr.Tab(label=name)
    for element in elements:
        tab.add_child(element)
    return tab


def create_tabs(*elements):
    tabs = gr.Tabs()
    for element in elements:
        tabs.add_child(element)
    return tabs


def create_blocks(*layouts):
    blocks = gr.Blocks()
    elements = list(layouts)
    while len(elements) > 0:
        element = elements.pop(0)
        if hasattr(element, "children"):
            elements += element.children
        blocks.blocks[element._id] = element

    for layout in layouts:
        blocks.add_child(layout)

    return blocks


# create some layouts


def create_pretrain_layout(pretrained_names: List[str], default_name: str):
    name = create_element(
        "dropdown",
        "Checkpoint",
        default=default_name,
        values=pretrained_names,
    )
    status = create_element("text", "Status", default="Stopped", interactive=False)
    start = create_element("button", "Start", variant="primary")
    stop = create_element("button", "Stop", variant="stop")
    layout = create_row(name, status, start, stop)
    return GenericOutputs(
        name=name,
        status=status,
        start=start,
        stop=stop,
        layout=layout,
    )


def create_controlnet_layout(
    controlnets: List[str], processes: List[str], num_controlnets: Optional[int] = 5
):
    def create_controlnet():
        input_image = create_element("image", "Input Image")
        output_image = create_element("image", "Output Image")
        checkpoint = create_element("dropdown", "Checkpoint", values=controlnets)
        guidance_scale = create_element(
            "slider", "Guidance Scale", min_value=0, max_value=2, step=0.1
        )
        process = create_element("radio", "Process", values=processes, default=None)
        layout = create_column(
            create_row(input_image, output_image),
            create_row(checkpoint, guidance_scale),
            create_row(process),
        )
        return (
            GenericOutputs(
                input_image=input_image,
                output_image=output_image,
                checkpoint=checkpoint,
                guidance_scale=guidance_scale,
                process=process,
            ),
            layout,
        )

    controlnet_groups = [create_controlnet() for _ in range(num_controlnets)]
    tabs = [
        create_tab(controlnet_group[-1], name=f"Net {i}")
        for i, controlnet_group in enumerate(controlnet_groups)
    ]
    layout = create_accordion(create_tabs(*tabs), name="ControlNets")
    return GenericOutputs(
        controlnets=[controlnet_group[0] for controlnet_group in controlnet_groups],
        layout=layout,
    )


def create_adapter_layout(
    adapters: List[str], processes: List[str], num_adapters: Optional[int] = 5
):
    def create_adapter():
        input_image = create_element("image", "Input Image")
        output_image = create_element("image", "Output Image")
        checkpoint = create_element("dropdown", "Checkpoint", values=adapters)
        guidance_scale = create_element(
            "slider", "Guidance Scale", min_value=0, max_value=2, step=0.1
        )
        process = create_element("radio", "Process", values=processes, default=None)
        layout = create_column(
            create_row(input_image, output_image),
            create_row(checkpoint, guidance_scale),
            create_row(process),
        )
        return (
            GenericOutputs(
                input_image=input_image,
                output_image=output_image,
                checkpoint=checkpoint,
                guidance_scale=guidance_scale,
                process=process,
            ),
            layout,
        )

    adapter_groups = [create_adapter() for _ in range(num_adapters)]
    tabs = [
        create_tab(controlnet_group[-1], name=f"Net {i}")
        for i, controlnet_group in enumerate(adapter_groups)
    ]
    layout = create_accordion(create_tabs(*tabs), name="Adapters")
    return GenericOutputs(
        adapters=[adapter_group[0] for adapter_group in adapter_groups],
        layout=layout,
    )


def create_lora_layout(
    loras: Union[List[str], Dict[str, List[str]]], num_loras: Optional[int] = 8
):
    def create_lora():
        checkpoint = create_element("dropdown", "Checkpoint", values=loras)
        weight = create_element(
            "slider", "Weight", default=1.0, min_value=0, max_value=3, step=0.1
        )
        alpha = create_element(
            "slider", "Alpha", default=32, min_value=0, max_value=128, step=1
        )
        url = create_element("text", "URL")
        file = create_element("file", "File")
        layout = create_column(
            create_row(checkpoint, create_column(weight, alpha)),
            create_row(url),
            create_row(file),
        )
        return (
            GenericOutputs(
                checkpoint=checkpoint, weight=weight, alpha=alpha, url=url, file=file
            ),
            layout,
        )

    lora_groups = [create_lora() for _ in range(num_loras)]
    tabs = [
        create_tab(lora_group[-1], name=f"LORA {i}")
        for i, lora_group in enumerate(lora_groups)
    ]
    layout = create_accordion(create_tabs(*tabs), name="LORAs")
    return GenericOutputs(
        loras=[lora_group[0] for lora_group in lora_groups], layout=layout
    )


def create_freeu_layout():
    s1 = create_element(
        "slider", "S1", default=0.9, min_value=0, max_value=10, step=0.1
    )
    s2 = create_element(
        "slider", "S2", default=0.2, min_value=0, max_value=10, step=0.1
    )
    b1 = create_element(
        "slider", "B1", default=1.2, min_value=0, max_value=10, step=0.1
    )
    b2 = create_element(
        "slider", "B2", default=1.4, min_value=0, max_value=10, step=0.1
    )
    layout = create_accordion(
        create_row(s1, s2), create_row(b1, b2), name="FreeU Params"
    )
    return GenericOutputs(s1=s1, s2=s2, b1=b1, b2=b2, layout=layout)



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.


import os
import io
import re
import fnmatch
import json
import signal
import time
import requests
import shutil
import logging
import tarfile
import tempfile
import torch
import safetensors
from filelock import FileLock
from contextlib import contextmanager
from torch.multiprocessing import spawn
from functools import partial, wraps
from hashlib import sha256
from pathlib import Path
from zipfile import ZipFile, is_zipfile
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
from transformers.utils.hub import is_remote_url, urlparse, http_get, http_user_agent
from huggingface_hub import HfFolder
from huggingface_hub.utils import hf_raise_for_status

from unitorch import is_offline_mode, get_cache_home
from unitorch.utils.decorators import replace
from unitorch.utils.functional import (
    pop_value,
    rpartial,
    truncate_sequence_pair,
    nested_dict_value,
)
from unitorch.utils.image_utils import (
    make_grid,
    resize_shortest_edge,
    image_list_to_tensor,
    numpy_to_pil,
)
from unitorch.utils.video_utils import tensor2vid
from unitorch.utils.import_utils import (
    is_deepspeed_available,
    is_megatron_available,
    is_fastapi_available,
    is_diffusers_available,
    is_xformers_available,
    is_opencv_available,
    is_torch_available,
    is_torch2_available,
    is_bitsandbytes_available,
    is_auto_gptq_available,
    is_onnxruntime_available,
)
from unitorch.utils.io import GenericWriter, IOProcess, PostProcess, GENERATE_FINISHED
from unitorch.utils.torch_utils import get_local_rank
from unitorch.utils.torch_utils import (
    DistributedSkipSampler,
    RandomSkipSampler,
    SequentialSkipSampler,
)


def url_to_filename(url: str, etag: Optional[str] = None) -> str:
    url_bytes = url.encode("utf-8")
    filename = sha256(url_bytes).hexdigest()

    if etag:
        etag_bytes = etag.encode("utf-8")
        filename += "." + sha256(etag_bytes).hexdigest()

    if url.endswith(".h5"):
        filename += ".h5"

    return filename


def get_from_cache(
    url: str,
    cache_dir=None,
    force_download=False,
    proxies=None,
    etag_timeout=10,
    resume_download=False,
    user_agent: Union[Dict, str, None] = None,
    use_auth_token: Union[bool, str, None] = None,
    local_files_only=False,
) -> Optional[str]:
    if cache_dir is None:
        cache_dir = get_cache_home()
    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    os.makedirs(cache_dir, exist_ok=True)

    headers = {"user-agent": http_user_agent(user_agent)}
    if isinstance(use_auth_token, str):
        headers["authorization"] = f"Bearer {use_auth_token}"
    elif use_auth_token:
        token = HfFolder.get_token()
        if token is None:
            raise EnvironmentError(
                "You specified use_auth_token=True, but a huggingface token was not found."
            )
        headers["authorization"] = f"Bearer {token}"

    url_to_download = url
    etag = None
    if not local_files_only:
        try:
            r = requests.head(
                url,
                headers=headers,
                allow_redirects=False,
                proxies=proxies,
                timeout=etag_timeout,
            )
            hf_raise_for_status(r)
            etag = r.headers.get("X-Linked-Etag") or r.headers.get("ETag")
            # We favor a custom header indicating the etag of the linked resource, and
            # we fallback to the regular etag header.
            # If we don't have any of those, raise an error.
            if etag is None:
                raise OSError(
                    "Distant resource does not have an ETag, we won't be able to reliably ensure reproducibility."
                )
            # In case of a redirect,
            # save an extra redirect on the request.get call,
            # and ensure we download the exact atomic version even if it changed
            # between the HEAD and the GET (unlikely, but hey).
            if 300 <= r.status_code <= 399:
                url_to_download = r.headers["Location"]
        except (requests.exceptions.SSLError, requests.exceptions.ProxyError):
            # Actually raise for those subclasses of ConnectionError
            raise
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
            # Otherwise, our Internet connection is down.
            # etag is None
            pass

    filename = url_to_filename(url, etag)

    # get cache path to put the file
    cache_path = os.path.join(cache_dir, filename)

    # etag is None == we don't have a connection or we passed local_files_only.
    # try to get the last downloaded one
    if etag is None:
        if os.path.exists(cache_path):
            return cache_path
        else:
            matching_files = [
                file
                for file in fnmatch.filter(
                    os.listdir(cache_dir), filename.split(".")[0] + ".*"
                )
                if not file.endswith(".json") and not file.endswith(".lock")
            ]
            if len(matching_files) > 0:
                return os.path.join(cache_dir, matching_files[-1])
            else:
                # If files cannot be found and local_files_only=True,
                # the models might've been found if local_files_only=False
                # Notify the user about that
                if local_files_only:
                    raise FileNotFoundError(
                        "Cannot find the requested files in the cached path and outgoing traffic has been"
                        " disabled. To enable model look-ups and downloads online, set 'local_files_only'"
                        " to False."
                    )
                else:
                    raise ValueError(
                        "Connection error, and we cannot find the requested files in the cached path."
                        " Please try again or make sure your Internet connection is on."
                    )

    # From now on, etag is not None.
    if os.path.exists(cache_path) and not force_download:
        return cache_path

    # Prevent parallel downloads of the same file with a lock.
    lock_path = cache_path + ".lock"
    with FileLock(lock_path):
        # If the download just completed while the lock was activated.
        if os.path.exists(cache_path) and not force_download:
            # Even if returning early like here, the lock will be released.
            return cache_path

        if resume_download:
            incomplete_path = cache_path + ".incomplete"

            @contextmanager
            def _resumable_file_manager() -> "io.BufferedWriter":
                with open(incomplete_path, "ab") as f:
                    yield f

            temp_file_manager = _resumable_file_manager
            if os.path.exists(incomplete_path):
                resume_size = os.stat(incomplete_path).st_size
            else:
                resume_size = 0
        else:
            temp_file_manager = partial(
                tempfile.NamedTemporaryFile, mode="wb", dir=cache_dir, delete=False
            )
            resume_size = 0

        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        with temp_file_manager() as temp_file:
            logging.debug(
                f"{url} not found in cache or force_download set to True, downloading to {temp_file.name}"
            )

            http_get(
                url_to_download,
                temp_file,
                proxies=proxies,
                resume_size=resume_size,
                headers=headers,
            )

        logging.debug(f"storing {url} in cache at {cache_path}")
        os.replace(temp_file.name, cache_path)

        # NamedTemporaryFile creates a file with hardwired 0600 perms (ignoring umask), so fixing it.
        umask = os.umask(0o666)
        os.umask(umask)
        os.chmod(cache_path, 0o666 & ~umask)

        logging.debug(f"creating metadata file for {cache_path}")
        meta = {"url": url, "etag": etag}
        meta_path = cache_path + ".json"
        with open(meta_path, "w") as meta_file:
            json.dump(meta, meta_file)

    return cache_path


def cached_path(
    url_or_filename: str,
    cache_dir: Optional[str] = None,
    force_download: Optional[bool] = False,
    proxies: Optional[Dict] = None,
    resume_download: Optional[bool] = False,
    user_agent: Union[Dict, str, None] = None,
    extract_compressed_file: Optional[bool] = False,
    force_extract: Optional[bool] = False,
    use_auth_token: Union[bool, str, None] = None,
    local_files_only: Optional[bool] = False,
) -> Optional[str]:
    """
    Retrieves a file from a given URL or local path and caches it for future use.

    Args:
        url_or_filename (str): URL or local path of the file.
        cache_dir (str, optional): Directory to store cached files. If not provided, the default cache directory will be used.
        force_download (bool, optional): Whether to force download the file even if it already exists in the cache. Defaults to False.
        proxies (dict, optional): Proxy configuration for downloading the file. Defaults to None.
        resume_download (bool, optional): Whether to resume a partially downloaded file. Defaults to False.
        user_agent (dict, str, None, optional): User agent configuration for the download request. Defaults to None.
        extract_compressed_file (bool, optional): Whether to extract compressed files (e.g., zip, tar) after downloading. Defaults to False.
        force_extract (bool, optional): Whether to force extract the file even if it has been extracted before. Defaults to False.
        use_auth_token (bool, str, None, optional): Authentication token to be used for downloading. Defaults to None.
        local_files_only (bool, optional): Whether to only use locally available files in offline mode. Defaults to False.

    Returns:
        Optional[str]: Path to the cached file or extracted directory, or None if the file is not found.

    Raises:
        EnvironmentError: If the file is not found or the URL or local path cannot be parsed.
        ValueError: If the URL or local path is in an unknown format.
    """
    if cache_dir is None:
        cache_dir = get_cache_home()

    if isinstance(url_or_filename, Path):
        url_or_filename = str(url_or_filename)

    if isinstance(cache_dir, Path):
        cache_dir = str(cache_dir)

    if is_offline_mode() and not local_files_only:
        logging.debug("Offline mode: forcing local_files_only=True")
        local_files_only = True

    if is_remote_url(url_or_filename):
        # URL, so get it from the cache (downloading if necessary)
        output_path = get_from_cache(
            url_or_filename,
            cache_dir=cache_dir,
            force_download=force_download,
            proxies=proxies,
            resume_download=resume_download,
            user_agent=user_agent,
            use_auth_token=use_auth_token,
            local_files_only=local_files_only,
        )
    elif os.path.exists(url_or_filename):
        # File, and it exists.
        output_path = url_or_filename
    elif urlparse(url_or_filename).scheme == "":
        # File, but it doesn't exist.
        raise EnvironmentError(f"file {url_or_filename} not found")
    else:
        # Something unknown
        raise ValueError(
            f"unable to parse {url_or_filename} as a URL or as a local path"
        )

    if extract_compressed_file:
        if not is_zipfile(output_path) and not tarfile.is_tarfile(output_path):
            return output_path

        # Path where we extract compressed archives
        # We avoid '.' in dir name and add "-extracted" at the end: "./model.zip" => "./model-zip-extracted/"
        output_dir, output_file = os.path.split(output_path)
        output_extract_dir_name = output_file.replace(".", "-") + "-extracted"
        output_path_extracted = os.path.join(output_dir, output_extract_dir_name)

        if (
            os.path.isdir(output_path_extracted)
            and os.listdir(output_path_extracted)
            and not force_extract
        ):
            return output_path_extracted

        # Prevent parallel extractions
        lock_path = output_path + ".lock"
        with FileLock(lock_path):
            shutil.rmtree(output_path_extracted, ignore_errors=True)
            os.makedirs(output_path_extracted)
            if is_zipfile(output_path):
                with ZipFile(output_path, "r") as zip_file:
                    zip_file.extractall(output_path_extracted)
                    zip_file.close()
            elif tarfile.is_tarfile(output_path):
                tar_file = tarfile.open(output_path)
                tar_file.extractall(output_path_extracted)
                tar_file.close()
            else:
                raise EnvironmentError(
                    f"Archive format of {output_path} could not be identified"
                )

        return output_path_extracted

    return output_path


def read_file(file, lines=False):
    result = open(file, "r").read()
    if lines:
        result = result.split("\n")
        if result[-1] == "":
            result = result[:-1]
    return result


def read_json_file(file):
    with open(file, "r") as f:
        return json.load(f)


def load_weight(
    path,
    replace_keys: Optional[Dict] = dict(),
    prefix_keys: Optional[Dict] = dict(),
):
    if isinstance(path, str):
        path = [path]

    state_dict = dict()
    for p in path:
        if p.endswith(".safetensors"):
            p = cached_path(p)
            state_dict = {**state_dict, **safetensors.torch.load_file(p)}
        else:
            p = cached_path(p)
            state_dict = {**state_dict, **torch.load(p, map_location="cpu")}

    results = dict()
    for key, value in list(state_dict.items()):
        for rkey, prefix in prefix_keys.items():
            if re.match(rkey, key):
                key = prefix + key
                break

        for rkey, nkey in replace_keys.items():
            key = re.sub(rkey, nkey, key)

        results[key] = value

    return results



# Copyright (c) FULIUCANSHENG.
# Licensed under the MIT License.

import os
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from unitorch.utils.io import GenericWriter
from unitorch.cli import add_default_section_for_init, register_writer


class WriterMixin:
    def from_pandas(self, df: pd.DataFrame):
        self.__pandas_dataframe__ = df

    def to_pandas(self):
        if hasattr(self, "__pandas_dataframe__"):
            return pd.DataFrame(self.__pandas_dataframe__)
        return pd.DataFrame()


class WriterOutputs:
    def __init__(
        self,
        process_outputs: pd.DataFrame,
    ):
        self.outputs = process_outputs
        self.schema = dict()

    def to_pandas(self):
        return self.outputs


@register_writer("core/writer/jsonl")
class GeneralJsonlWriter(GenericWriter):
    """Class for writing data in JSONL format."""

    def __init__(
        self,
        output_file: str,
        nrows_per_sample: Optional[int] = None,
        header: Optional[bool] = None,
        columns: Optional[List[str]] = None,
    ):
        """
        Initialize GeneralJsonlWriter.

        Args:
            output_file (str): The path to the output file.
            nrows_per_sample (int, optional): The number of rows per sample. Defaults to None.
            header (bool, optional): Whether to include a header in the output file. Defaults to None.
            columns (List[str], optional): The list of columns to include in the output file. Defaults to None.
        """
        self.header = header
        self.columns = columns
        self.skip_n_samples = (
            0
            if nrows_per_sample is None or not os.path.exists(output_file)
            else sum(1 for _ in open(output_file)) // nrows_per_sample
        )
        if self.skip_n_samples == 0:
            self.output_file = open(output_file, "w", encoding="utf-8")
        else:
            self.output_file = open(output_file, "a", encoding="utf-8")

    @classmethod
    @add_default_section_for_init("core/writer/jsonl")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of GeneralJsonlWriter from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            GeneralJsonlWriter: An instance of GeneralJsonlWriter.
        """
        pass

    def process_start(self, outputs: WriterOutputs):
        """
        Process the start of the writing process.

        Args:
            outputs (WriterOutputs): The writer outputs.
        """
        dataframe = outputs.to_pandas()
        if self.columns is not None:
            columns = set(dataframe.columns)
            dataframe = dataframe[[h for h in self.columns if h in columns]]
        string = dataframe.to_json(orient="records", lines=True)
        self.output_file.write(string)
        self.output_file.flush()

    def process_end(self):
        """Process the end of the writing process."""
        self.output_file.close()

    def process_chunk(self, outputs: WriterOutputs):
        """
        Process a chunk of data during the writing process.

        Args:
            outputs (WriterOutputs): The writer outputs.
        """
        dataframe = outputs.to_pandas()
        if self.columns is not None:
            columns = set(dataframe.columns)
            dataframe = dataframe[[h for h in self.columns if h in columns]]
        string = dataframe.to_json(orient="records", lines=True)
        self.output_file.write(string)
        self.output_file.flush()


@register_writer("core/writer/csv")
class GeneralCsvWriter(GenericWriter):
    """Class for writing data in CSV format."""

    def __init__(
        self,
        output_file: str,
        nrows_per_sample: Optional[int] = None,
        header: Optional[bool] = None,
        columns: Optional[List[str]] = None,
        sep: Optional[str] = "\t",
        quoting: Optional[int] = 3,
        escapechar: Optional[str] = None,
    ):
        """
        Initialize GeneralCsvWriter.

        Args:
            output_file (str): The path to the output file.
            nrows_per_sample (int, optional): The number of rows per sample. Defaults to None.
            header (bool, optional): Whether to include a header in the output file. Defaults to None.
            columns (List[str], optional): The list of columns to include in the output file. Defaults to None.
            sep (str, optional): The separator for the CSV file. Defaults to "\t".
            quoting (int, optional): The quoting style for the CSV file. Defaults to 3.
            escapechar (str, optional): The escape character for the CSV file. Defaults to None.
        """
        self.header = header
        self.columns = columns
        self.sep = sep
        self.quoting = quoting
        self.escapechar = escapechar
        has_header = int(header is True)
        self.skip_n_samples = (
            0
            if nrows_per_sample is None or not os.path.exists(output_file)
            else (sum(1 for _ in open(output_file)) - has_header) // nrows_per_sample
        )
        if self.skip_n_samples == 0:
            self.output_file = open(output_file, "w", encoding="utf-8")
        else:
            self.output_file = open(output_file, "a", encoding="utf-8")

    @classmethod
    @add_default_section_for_init("core/writer/csv")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of GeneralCsvWriter from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            GeneralCsvWriter: An instance of GeneralCsvWriter.
        """
        pass

    def process_start(self, outputs: WriterOutputs):
        """
        Process the start of the writing process.

        Args:
            outputs (WriterOutputs): The writer outputs.
        """
        dataframe = outputs.to_pandas()
        if self.columns is not None:
            columns = set(dataframe.columns)
            dataframe = dataframe[[h for h in self.columns if h in columns]]
        string = dataframe.to_csv(
            index=False,
            sep=self.sep,
            quoting=self.quoting,
            header=self.header and self.skip_n_samples == 0,
            escapechar=self.escapechar,
        )
        self.output_file.write(string)
        self.output_file.flush()

    def process_end(self):
        """Process the end of the writing process."""
        self.output_file.close()

    def process_chunk(self, outputs: WriterOutputs):
        """
        Process a chunk of data during the writing process.

        Args:
            outputs (WriterOutputs): The writer outputs.
        """
        dataframe = outputs.to_pandas()
        if self.columns is not None:
            columns = set(dataframe.columns)
            dataframe = dataframe[[h for h in self.columns if h in columns]]
        string = dataframe.to_csv(
            index=False,
            sep=self.sep,
            quoting=self.quoting,
            header=False,
            escapechar=self.escapechar,
        )
        self.output_file.write(string)
        self.output_file.flush()


@register_writer("core/writer/parquet")
class GeneralParquetWriter(GenericWriter):
    def __init__(
        self,
        output_file: str,
        nrows_per_sample: Optional[int] = None,
        columns: Optional[List[str]] = None,
        schema: Optional[str] = None,
        compression: Optional[str] = "snappy",
    ):
        """
        Initialize GeneralParquetWriter.

        Args:
            output_file (str): The path to the output file.
            nrows_per_sample (int, optional): The number of rows per sample. Defaults to None.
            columns (List[str], optional): The list of columns to include in the output file. Defaults to None.
            schema (str, optional): The Parquet schema in string format. Defaults to None.
            compression (str, optional): The compression algorithm to use. Defaults to "snappy".
        """
        self.columns = columns
        self.skip_n_samples = 0
        self.output_file = output_file
        self.pq_writer = None
        self.pq_schema = None if schema is None else eval(schema)
        self.compression = compression

    @classmethod
    @add_default_section_for_init("core/writer/parquet")
    def from_core_configure(cls, config, **kwargs):
        """
        Create an instance of GeneralParquetWriter from a core configuration.

        Args:
            config: The core configuration.
            **kwargs: Additional keyword arguments.

        Returns:
            GeneralParquetWriter: An instance of GeneralParquetWriter.
        """
        pass

    def process_start(self, outputs: WriterOutputs):
        """
        Process the start of the writing process.

        Args:
            outputs (WriterOutputs): The writer outputs.
        """
        dataframe = outputs.to_pandas()
        if self.columns is not None:
            columns = set(dataframe.columns)
            dataframe = dataframe[[h for h in self.columns if h in columns]]
        pa_table = pa.Table.from_pandas(dataframe)
        if self.pq_schema is None:
            self.pq_schema = pa_table.schema
        pa_table = pa.Table.from_pandas(dataframe, schema=self.pq_schema)
        self.pq_writer = pq.ParquetWriter(
            self.output_file,
            self.pq_schema,
            version="1.0",
            use_dictionary=False,
            flavor="spark",
            compression=self.compression,
            use_compliant_nested_type=True,
        )
        self.pq_writer.write_table(pa_table)

    def process_end(self):
        """Process the end of the writing process."""
        self.pq_writer.close()

    def process_chunk(self, outputs: WriterOutputs):
        """
        Process a chunk of data during the writing process.

        Args:
            outputs (WriterOutputs): The writer outputs.
        """
        assert self.pq_writer is not None
        dataframe = outputs.to_pandas()
        if self.columns is not None:
            columns = set(dataframe.columns)
            dataframe = dataframe[[h for h in self.columns if h in columns]]
        pa_table = pa.Table.from_pandas(dataframe, schema=self.pq_schema)
        self.pq_writer.write_table(pa_table)


# more writers
